from .configuration import Configuration


class Attck(object):

    """Interface to all MITRE ATT&CK frameworks.

    Currently, this class enables access to the Enterprise & PRE-ATT&CK
    frameworks with others coming soon.  To acccess each framework, use
    the following properties

        * enterprise
        * preattack

    This interface enables you to retrieve all properties within each
    item in the MITRE ATT&CK Enterprise Framework.

    The following categorical items can be accessed using this class:

        1. Tactics (Tactics are the phases defined by MITRE ATT&CK)
        2. Techniques (Techniques are the individual actions which can
            accomplish a tactic)
        3. Mitigations (Mitigations are recommendations to prevent or
            protect against a technique)
        4. Actors (Actors or Groups are identified malicious
            actors/groups which have been identified and documented by
            MITRE & third-parties)
        5. Tools (Tools are software used to perform techniques)
        6. Malwares (Malwares are specific pieces of malware used by
            actors (or in general) to accomplish a technique)

    You can also search the external dataset for external commands that
    are similar using the `search_commands` method.

        .. code-block:: python

            from pyattck import Attck

            attck = Attck()

            for search in attck.enterprise.search_commands('powershell'):
                print(search['technique'])
                print(search['reason_for_match'])

    You can access additional datasets related to a technique. 
    These datasets are [documented here](https://github.com/swimlane/pyattck-data).

    Example:
        Once an Attck object is instantiated, you can access each object
        type as a list of objects (e.g. techniques, tactics, actors, etc.)

        You can iterate over each object list and access specific properties
        and relationship properties of each.

        The following relationship properties are accessible:
            1. Actors
                1. Tools used by the Actor or Group
                2. Malware used by the Actor or Group
                3. Techniques this Actor or Group uses
            2. Malwares
                1. Actor or Group(s) using this malware
                2. Techniques this malware is used with
            3. Mitigations
                1. Techniques related to a specific set of mitigation suggestions
            4. Tactics
                1. Techniques found in a specific Tactic (phase)
            5. Techniques
                1. Tactics a technique is found in
                2. Mitigation suggestions for a given technique
                3. Actor or Group(s) identified as using this technique
            6. Tools
                1. Techniques that the specified tool is used within
                2. Actor or Group(s) using a specified tool

            1. To iterate over a list, do the following:

            .. code-block:: python

               from pyattck import Attck

               attck = Attck()

               for technique in attck.enterprise.techniques:
                   print(technique.id)
                   print(technique.name)
                   print(technique.description)
                   # etc.
               for mitigation in attck.enterprise.mitigations:
                   print(mitigation.id)
                   print(mitigation.name)
                   print(mitigation.description)
                   # etc.

            2. To access relationship properties, do the following:

            .. code-block:: python

               from pyattck import Attck

               attck = Attck()

               for technique in attck.enterprise.techniques:
                   print(technique.id)
                   print(technique.name)
                   print(technique.description)
                   # etc.

                   for actor in technique.enterprise.actors:
                       print(actor.id)
                       print(actor.name)
                       print(actor.description)
                       # etc.

               for mitigation in attck.enterprise.mitigations:
                   print(mitigation.id)
                   print(mitigation.name)
                   print(mitigation.description)
                   # etc.

                   for technique in mitigation.enterprise.techniques:
                       print(technique.name)
                       print(technique.description)
                       # etc.

    Arguments:
        nested_subtechniques (bool, optional): Whether not to iterate over nested subtechniques. Defaults to True.
        use_config (bool, optional): Specifies if a configuration file should be used or not.  Defaults to False.
        save_config (bool, optional): Specifies if pyattck should save a configuration file based on the provided 
                                      values.  Defaults to False.
        config_file_path (str, optional): Path to a yaml configuration file which contains two key value pairs. 
                                          Defaults to '~/pyattck/config.yml'.
        data_path (str, optional): Path to store the external data locally on your system. Defaults to '~/pyattck/data'.
        enterprise_attck_json (str, optional): A URL or local file path to the MITRE ATT&CK Json file. 
                                               Defaults to https://raw.githubusercontent.com/mitre/cti/master/enterprise-attack/enterprise-attack.json.
        pre_attck_json (str, optional): A URL or local file path to the MITRE Pre-ATT&CK Json file. 
                                        Defaults to https://raw.githubusercontent.com/mitre/cti/master/pre-attack/pre-attack.json.
        mobile_attck_json (str, optional): A URL or local file path to the MITRE Mobile ATT&CK Json file. 
                                           Defaults to https://raw.githubusercontent.com/mitre/cti/master/mobile-attack/mobile-attack.json.
        ics_attck_json (str, optional): A URL or local file path to the MITRE ICS ATT&CK JSON file.
                                           Defaults to https://raw.githubusercontent.com/mitre/cti/master/ics-attack/ics-attack.json.
        nist_controls_json (str, optional): A URL or local file path to the NIST Controls Json file. 
                                            Defaults to https://raw.githubusercontent.com/center-for-threat-informed-defense/attack-control-framework-mappings/master/frameworks/ATT%26CK-v9.0/nist800-53-r5/stix/nist800-53-r5-controls.json.
        generated_attck_json (str, optional): A URL or local file path to the Generated MITRE ATT&CK Json file. 
                                              Defaults to https://swimlane-pyattck.s3.us-west-2.amazonaws.com/generated_attck_data.json.
        generated_nist_json (str, optional): A URL or local file path to the Generated NIST Controls Mapping Json file. 
                                             Defaults to https://swimlane-pyattck.s3.us-west-2.amazonaws.com/attck_to_nist_controls.json.
        kwargs (dict, optional): Provided kwargs will be passed to any HTTP requests using the Requests library. 
                                 Defaults to None.

    Returns:
        [Attck]: Returns a Attck object that contains all data from MITRE ATT&CK Frameworks
    """

    def __init__(
        self,
        nested_subtechniques=True,
        use_config=False,
        save_config=False,
        config_file_path='~/pyattck/config.yml',
        data_path='~/pyattck/data',
        enterprise_attck_json="https://raw.githubusercontent.com/mitre/cti/master/enterprise-attack/enterprise-attack.json",
        pre_attck_json="https://raw.githubusercontent.com/mitre/cti/master/pre-attack/pre-attack.json",
        mobile_attck_json="https://raw.githubusercontent.com/mitre/cti/master/mobile-attack/mobile-attack.json",
        ics_attck_json="https://raw.githubusercontent.com/mitre/cti/master/ics-attack/ics-attack.json",
        nist_controls_json="https://raw.githubusercontent.com/center-for-threat-informed-defense/attack-control-framework-mappings/master/frameworks/ATT%26CK-v9.0/nist800-53-r5/stix/nist800-53-r5-controls.json",
        generated_attck_json="https://swimlane-pyattck.s3.us-west-2.amazonaws.com/generated_attck_data.json",
        generated_nist_json="https://swimlane-pyattck.s3.us-west-2.amazonaws.com/attck_to_nist_controls.json",
        **kwargs
        ):
        """
        The main entry point for pyattck.

        When instantiating an Attck object you can specify if you want the
        new subtechniques to be nested underneath their parent techniques
        or not.

        Setting nested_subtechniques to False will result in all techniques
        accessible under the techniques property. If using the default value
        of True, subtechniques will be accessible underneath
        technique.subtechniques.

        When instantiating an Attck object you can access either the
        Enterprise, PRE-ATT&CK, or Mobile MITRE Frameworks.  Specify
        one of the following properties to access the frameworks specific
        data:

            * enterprise
            * preattack
            * mobile
            * ics

        You can specify an alternate location of a local copy of the
        following objects:

            1. config_file_path = Path to a yaml configuration file
                                  which contains two key value pairs
                Example content:

                    config_file_path: /Users/user.name/pyattck/config.yml
                    data_path: /Users/user.name/pyattck/data
                    enterprise_attck_json: https://raw.githubusercontent.com/mitre/cti/master/enterprise-attack/enterprise-attack.json
                    generated_attck_json: https://swimlane-pyattck.s3.us-west-2.amazonaws.com/generated_attck_data.json
                    generated_nist_json: https://swimlane-pyattck.s3.us-west-2.amazonaws.com/attck_to_nist_controls.json
                    mobile_attck_json: https://raw.githubusercontent.com/mitre/cti/master/mobile-attack/mobile-attack.json
                    ics_attck_json: https://raw.githubusercontent.com/mitre/cti/master/ics-attack/ics-attack.json
                    nist_controls_json: https://raw.githubusercontent.com/center-for-threat-informed-defense/attack-control-framework-mappings/master/frameworks/ATT%26CK-v9.0/nist800-53-r5/stix/nist800-53-r5-controls.json
                    pre_attck_json: https://raw.githubusercontent.com/mitre/cti/master/pre-attack/pre-attack.json


            2. data_path = The path to hold the external data locally on your system.
                           The default is your user home path.

        Args:
            nested_subtechniques (bool, optional): Whether not to iterate over nested subtechniques. Defaults to True.
            use_config (bool, optional): Specifies if a configuration file should be used or not.  Defaults to False.
            save_config (bool, optional): Specifies if pyattck should save a configuration file based on the 
                                          provided values.  Defaults to False.
            config_file_path (str, optional): Path to a yaml configuration file which contains two key value pairs. 
                                              Defaults to '~/pyattck/config.yml'.
            data_path (str, optional): Path to store the external data locally on your system.  
                                       Defaults to '~/pyattck/data'.
            enterprise_attck_json (str, optional): A URL or local file path to the MITRE ATT&CK Json file. 
                                                   Defaults to https://raw.githubusercontent.com/mitre/cti/master/enterprise-attack/enterprise-attack.json.
            pre_attck_json (str, optional): A URL or local file path to the MITRE Pre-ATT&CK Json file. 
                                            Defaults to https://raw.githubusercontent.com/mitre/cti/master/pre-attack/pre-attack.json.
            mobile_attck_json (str, optional): A URL or local file path to the MITRE Mobile ATT&CK Json file. 
                                               Defaults to https://raw.githubusercontent.com/mitre/cti/master/mobile-attack/mobile-attack.json.
            ics_attck_json (str, optional): A URL or local file path to the MITRE ICS ATT&CK JSON file.
                                           Defaults to https://raw.githubusercontent.com/mitre/cti/master/ics-attack/ics-attack.json.
            nist_controls_json (str, optional): A URL or local file path to the NIST Controls Json file. 
                                                Defaults to https://raw.githubusercontent.com/center-for-threat-informed-defense/attack-control-framework-mappings/master/frameworks/ATT%26CK-v9.0/nist800-53-r5/stix/nist800-53-r5-controls.json
            generated_attck_json (str, optional): A URL or local file path to the Generated MITRE ATT&CK Json file. 
                                                  Defaults to https://swimlane-pyattck.s3.us-west-2.amazonaws.com/generated_attck_data.json.
            generated_nist_json (str, optional): A URL or local file path to the Generated NIST Controls Mapping 
                                                 Json file. 
                                                 Defaults to https://swimlane-pyattck.s3.us-west-2.amazonaws.com/attck_to_nist_controls.json.
            kwargs (dict, optional): Provided kwargs will be passed to any HTTP requests using the Requests library. 
                                     Defaults to None.
        """
        self.__nested_subtechniques = nested_subtechniques
        Configuration.use_config = use_config
        Configuration.save_config = save_config
        Configuration.config_file_path = config_file_path
        Configuration.enterprise_attck_json = enterprise_attck_json
        Configuration.pre_attck_json = pre_attck_json
        Configuration.mobile_attck_json = mobile_attck_json
        Configuration.ics_attck_json = ics_attck_json
        Configuration.nist_controls_json = nist_controls_json
        Configuration.generated_attck_json = generated_attck_json
        Configuration.generated_nist_json = generated_nist_json
        Configuration.data_path = data_path
        Configuration.requests_kwargs = kwargs

    @property
    def enterprise(self):
        """
        Retrieve objects from the Enterprise MITRE ATT&CK Framework and
        additional generated data which provides additional context

        Returns:
            Enterprise: Returns an Enterprise object
        """
        from .enterprise.enterprise import Enterprise
        return Enterprise(nested_subtechniques=self.__nested_subtechniques)

    @property
    def preattack(self):
        """
        Retrieve objects from the MITRE PRE-ATT&CK Framework

        Returns:
            PreAttack: Returns an PreAttack object
        """
        from .preattck.preattck import PreAttck
        return PreAttck()

    @property
    def mobile(self):
        """
        Retrieve objects from the MITRE Mobile ATT&CK Framework

        Returns:
            PreAttack: Returns an MobileAttack object
        """
        from .mobile.mobileattck import MobileAttck
        return MobileAttck()

    @property
    def ics(self):
        """
        Retrieve objects from the MITRE ICS ATT&CK Framework

        Returns:
            PreAttack: Returns an ICSAttck object
        """
        from .ics.icsattck import ICSAttck
        return ICSAttck()

    def update(self) -> bool:
        return True if Configuration._save_json_data(force=True) else False



from .attckobject import AttckObject


class AttckTools(AttckObject):

    """Enterprise MITRE ATT&CK Tool object.

    A child class of AttckObject

    Creates objects which have been categorized as tools or software which have
    been categorized as software used in attacks

    You can also access external data properties. The following properties are
    generated using external data:

        1. additional_names
        2. attribution_links
        3. additional_comments
        4. family

    You can retrieve the entire dataset using the `external_dataset` property.

    You can also access external data properties from the C2 Matrix project.
    The following properties are generated using C2 Matrix external data:

        - HTTP
        - Implementation
        - Custom Profile
        - DomainFront
        - Multi-User
        - SMB
        - Kill Date
        - macOS
        - GitHub
        - Key Exchange
        - Chaining
        - Price
        - TCP
        - Proxy Aware
        - HTTP3
        - HTTP2
        - Date
        - Evaluator
        - Working Hours
        - Slack
        - FTP
        - Version Reviewed
        - Logging
        - Name
        - License
        - Windows
        - Stego
        - Notes
        - Server
        - Actively Maint.
        - Dashboard
        - DNS
        - Popular Site
        - ICMP
        - IMAP
        - DoH
        - Jitter
        - How-To
        - ATT&CK Mapping
        - Kali
        - Twitter
        - MAPI
        - Site
        - Agent
        - API
        - UI
        - Linux

    You can retrieve the entire dataset using the `c2_data` property.

    Example:
        You can iterate over an `tools` list and access specific properties
        and relationship properties.

        The following relationship properties are accessible:
                1. techniques
                2. actors

            1. To iterate over an `tools` list, do the following:

            .. code-block:: python

            from pyattck import Attck

            attck = Attck()

            for tool in attck.enterprise.tools:
                print(tool.id)
                print(tool.name)
                print(tool.description)
                # etc.

            2. To access relationship properties, do the following:

            .. code-block:: python

            from pyattck import Attck

            attck = Attck()

            for tool in attck.enterprise.tools:
                print(tool.id)
                print(tool.name)
                print(tool.description)
                # etc.

                for technique in tool.techniques:
                    print(technique.name)
                    print(technique.description)
                    # etc.

    Arguments:
        attck_obj (json) -- Takes the raw MITRE ATT&CK Json object
        AttckObject (dict) -- Takes the MITRE ATT&CK Json object as a kwargs values
    """

    def __init__(self, attck_obj = None, **kwargs):
        """
        This class represents a Tool as defined by the
        Enterprise MITRE ATT&CK framework.

        Keyword Arguments:
            attck_obj {json} -- A Enterprise MITRE ATT&CK Framework
                                json object (default: {None})

        Raises:
            GeneratedDatasetException: Raised an exception when unable to access
                                       or process the external generated dataset.
        """
        super(AttckTools, self).__init__(**kwargs)
        self.__attck_obj = attck_obj
        self.id = self._set_id(kwargs)
        self.name = self._set_attribute(kwargs, 'name')
        self.description = self._set_attribute(kwargs, 'description')
        self.reference = self._set_reference(kwargs)
        self.created = self._set_attribute(kwargs, 'created')
        self.modified = self._set_attribute(kwargs, 'modified')
        self.stix = self._set_attribute(kwargs, 'id')
        self.type = self._set_attribute(kwargs, 'type')
        self.wiki = self._set_wiki(kwargs)
        self.contributor = self._set_attribute(kwargs, 'contributor')
        self.set_relationships(self.__attck_obj)
        self.c2_data = self.__get_c2_dataset()
        self.external_dataset =  self.__get_tools_dataset()

    def __get_tools_dataset(self):
        return_list = []
        self.additional_names = set()
        self.attribution_links = set()
        self.additional_comments = set()
        self.family = set()
        for tool in AttckObject.generated_attck_json['tools'].get('tools'):
            if tool.get('names'):
                if self.name.lower() in [x.lower() for x in tool.get('names')] or hasattr(self, 'alias') and any(x.lower() in tool.get('names') for x in self.alias):
                    self.additional_names.update(tool.get('names'))
                    if tool.get('links'):
                        self.attribution_links.update(tool['links'])
                    if tool.get('family'):
                        self.family.update(tool['family'])
                    if tool.get('comments'):
                        self.additional_comments.update(tool['comments'])
                    return_list.append(tool)
        self.additional_names = list(self.additional_names)
        self.attribution_links = list(self.attribution_links)
        self.additional_comments = list(self.additional_comments)
        self.family = list(self.family)
        return return_list

    def __get_c2_dataset(self):
        return_dict = {}
        for k,v in AttckObject.generated_attck_json['c2_data'].items():
            if self.name.lower() == k.lower():
                return_dict[k] = v
                for key,val in v.items():
                    try:
                        setattr(self, key, val)
                    except:
                        setattr(self, 'c2_{}'.format(key), val)
            if self.alias:
                for item in self.alias:
                    if item.lower() == k.lower():
                        return_dict[k] = v
                        for key,val in v.items():
                            try:
                                setattr(self, key, val)
                            except:
                                setattr(self, 'c2_{}_{}'.format(item.lower(), key), val)
        return return_dict if return_dict else None

    @property
    def techniques(self):
        """
        Returns all technique objects that a tool object is associated with

        Returns:
            [list] -- A list of technique objects defined within the
                      Enterprise MITRE ATT&CK Framework
        """
        from .technique import AttckTechnique
        return_list = []
        item_dict = {}
        for item in self.__attck_obj['objects']:
            if 'type' in item:
                if item['type'] == 'attack-pattern':
                    item_dict[item['id']] = item
        if self._RELATIONSHIPS.get(self.stix):
            for item in self._RELATIONSHIPS[self.stix]:
                if item in item_dict:
                    return_list.append(AttckTechnique(attck_obj=self.__attck_obj, **item_dict[item]))
        return return_list

    @property
    def actors(self):
        """
        Returns all actor objects that use a tool

        Returns:
            [list] -- A list of actor objects defined within the
                      Enterprise MITRE ATT&CK Framework
        """
        from .actor import AttckActor
        return_list = []
        item_dict = {}
        for item in self.__attck_obj['objects']:
            if 'type' in item:
                if item['type'] == 'intrusion-set':
                    item_dict[item['id']] = item
        if self._RELATIONSHIPS.get(self.stix):
            for item in self._RELATIONSHIPS[self.stix]:
                if item in item_dict:
                    return_list.append(AttckActor(attck_obj=self.__attck_obj, **item_dict[item]))
        return return_list



import os

from pyfiglet import Figlet

ASCII_CHARS = ['.',',',':',';','+','*','?','%','S','#','@']
ASCII_CHARS = ASCII_CHARS[::-1]

class Logo(object):

    """Generate or retrieve a logo for a given MITRE ATT&CK Actor or Group name

    Returns:
        str: The path to a image logo or string containing a generated ASCII text
             based on the Actor or Group name
    """

    __DATA_PATH = os.path.abspath(os.path.join(os.path.dirname(os.path.dirname(__file__)), 'data', 'actors'))

    def __init__(self, actor_name):
        """
        Retrieves or generates a image based on a provided malicious actor name.

        Args:
            actor_name ([str]): A malicious actors name
        """
        self.image_path = None
        self.__check_for_logo(actor_name)

    def get_ascii(self):
        """
        Retrieve a ASCII string of an Actor or Group by name

        Returns:
            str: ASCII string of a generated logo
        """
        return self.__image

    def get_image(self):
        """
        Retrieve a image logo path of an Actor or Group by name

        Returns:
            str: The path to the image logo of an Actor or Group by name
        """
        if self.image_path:
            return self.image_path
        return None

    def __check_for_logo(self, actor_name):
        if os.path.exists('{}/{}.png'.format(self.__DATA_PATH, actor_name)):
            self.convert_image('{}/{}.png'.format(self.__DATA_PATH, actor_name))
            self.image_path = os.path.abspath('{}/{}.png'.format(self.__DATA_PATH, actor_name))
        else:
            custom_fig = Figlet(font='graffiti')
            self.__image = custom_fig.renderText(actor_name)

    def convert_image(self, path):
        image = None
        try:
            from PIL import Image
            image = Image.open(path)
        except:
            try:
                import Image
                image = Image.open(path)
            except:
                pass
            pass
        if image:
            self.__image = self.__process(image)
        else:
            self.__image = None

    def __resize(self, image, new_width=100):
        """
        method resize():
            - takes as parameters the image, and the final width
            - resizes the image into the final width while maintaining aspect ratio
        """
        (old_width, old_height) = image.size
        aspect_ratio = float(old_height)/float(old_width)
        new_height = int(aspect_ratio * new_width)
        new_dim = (new_width, new_height)
        new_image = image.resize(new_dim)
        return new_image

    def __gray_scale(self, image):
        """
        method grayscalify():
            - takes an image as a parameter
            - returns the grayscale version of image
        """
        return image.convert('L')

    def __modify(self, image, buckets=25):
        """
        method modify():
            - replaces every pixel with a character whose intensity is similar
        """
        initial_pixels = list(image.getdata())
        new_pixels = [ASCII_CHARS[pixel_value//buckets] for pixel_value in initial_pixels]
        return ''.join(new_pixels)

    def __process(self, image, new_width=100):
        """
        method do():
            - does all the work by calling all the above functions
        """
        image = self.__resize(image)
        image = self.__gray_scale(image)
        pixels = self.__modify(image)
        len_pixels = len(pixels)
        # Construct the image from the character list
        new_image = [pixels[index:index+new_width] for index in range(0, len_pixels, new_width)]
        return '\n'.join(new_image)



from .mobileattckobject import MobileAttckObject


class MobileAttckTools(MobileAttckObject):

    """Mobile MITRE ATT&CK Tool object.

    A child class of MobileAttckObject

    Creates objects which have been categorized as tools or software
    which have been categorized as software used in attacks

    You can also access external data properties. The following
    properties are generated using external data:

        1. additional_names
        2. attribution_links
        3. additional_comments
        4. family

    You can retrieve the entire dataset using the `external_dataset` property.

    You can also access external data properties from the C2 Matrix project.
    The following properties are generated using C2 Matrix external data:

        - HTTP
        - Implementation
        - Custom Profile
        - DomainFront
        - Multi-User
        - SMB
        - Kill Date
        - macOS
        - GitHub
        - Key Exchange
        - Chaining
        - Price
        - TCP
        - Proxy Aware
        - HTTP3
        - HTTP2
        - Date
        - Evaluator
        - Working Hours
        - Slack
        - FTP
        - Version Reviewed
        - Logging
        - Name
        - License
        - Windows
        - Stego
        - Notes
        - Server
        - Actively Maint.
        - Dashboard
        - DNS
        - Popular Site
        - ICMP
        - IMAP
        - DoH
        - Jitter
        - How-To
        - ATT&CK Mapping
        - Kali
        - Twitter
        - MAPI
        - Site
        - Agent
        - API
        - UI
        - Linux

    You can retrieve the entire dataset using the `c2_data` property.

    Example:
        You can iterate over an `tools` list and access specific properties
        and relationship properties.

        The following relationship properties are accessible:
                1. techniques
                2. actors

            1. To iterate over an `tools` list, do the following:

            .. code-block:: python

            from pyattck import Attck

            attck = Attck()

            for tool in attck.mobile.tools:
                print(tool.id)
                print(tool.name)
                print(tool.description)
                # etc.

            2. To access relationship properties, do the following:

            .. code-block:: python

            from pyattck import Attck

            attck = Attck()

            for tool in attck.mobile.tools:
                print(tool.id)
                print(tool.name)
                print(tool.description)
                # etc.

                for technique in tool.techniques:
                    print(technique.name)
                    print(technique.description)
                    # etc.
    """

    def __init__(self, mobile_attck_obj = None, **kwargs):
        """
        Creates an MobileAttckTools object.  
        The MobileAttckTools object is based on software which have been
        categorized as software used in attacks

        Arguments:
            mobile_attck_obj (json) -- Takes the raw MITRE Mobile
                                       ATT&CK Json object
            AttckObject (dict) -- Takes the MITRE Mobile ATT&CK Json object
                                  as a kwargs values
        """
        super(MobileAttckTools, self).__init__(**kwargs)
        self.__mobile_attck_obj = mobile_attck_obj
        self.external_reference = self._set_reference(kwargs)
        self.platforms = self._set_list_items(kwargs, 'x_mitre_platforms')
        self.version = self._set_attribute(kwargs, 'x_mitre_version')
        self.created_by_ref = self._set_attribute(kwargs, 'created_by_ref')
        self.contributor = self._set_list_items(kwargs, 'x_mitre_contributors')
        self.labels = self._set_list_items(kwargs, 'labels')
        self.old_attack_id = self._set_attribute(kwargs, 'x_mitre_old_attack_id')
        self.stix = self._set_attribute(kwargs, 'id')
        self.wiki = self._set_wiki(kwargs)
        self.set_relationships(self.__mobile_attck_obj)
        self.c2_data = self.__get_c2_dataset()
        self.external_dataset =  self.__get_tools_dataset()

    def __get_tools_dataset(self):
        return_list = []
        self.additional_names = set()
        self.attribution_links = set()
        self.additional_comments = set()
        self.family = set()
        for tool in MobileAttckObject.generated_attck_json['tools'].get('tools'):
            if tool.get('names'):
                if self.name.lower() in [x.lower() for x in tool.get('names')] or hasattr(self, 'alias') and any(x.lower() in tool.get('names') for x in self.alias):
                    self.additional_names.update(tool.get('names'))
                    if tool.get('links'):
                        self.attribution_links.update(tool['links'])
                    if tool.get('family'):
                        self.family.update(tool['family'])
                    if tool.get('comments'):
                        self.additional_comments.update(tool['comments'])
                    return_list.append(tool)
        self.additional_names = list(self.additional_names)
        self.attribution_links = list(self.attribution_links)
        self.additional_comments = list(self.additional_comments)
        self.family = list(self.family)
        return return_list

    def __get_c2_dataset(self):
        return_dict = {}
        for k,v in MobileAttckObject.generated_attck_json['c2_data'].items():
            if self.name.lower() == k.lower():
                return_dict[k] = v
                for key,val in v.items():
                    try:
                        setattr(self, key, val)
                    except:
                        setattr(self, 'c2_{}'.format(key), val)
            if self.alias:
                for item in self.alias:
                    if item and item.lower() == k.lower():
                        return_dict[k] = v
                        for key,val in v.items():
                            try:
                                setattr(self, key, val)
                            except:
                                setattr(self, 'c2_{}_{}'.format(item.lower(), key), val)
        return return_dict if return_dict else None

    @property
    def techniques(self):
        """
        Accessing techniques which have been seen using a specific tool
        as part of the MITRE Mobile ATT&CK Framework

        Returns:
            list: Returns all technique objects as a list that are associated
                  with a tool
        """
        from .technique import MobileAttckTechnique
        return_list = []
        item_dict = {}
        for item in self.__mobile_attck_obj['objects']:
            if 'type' in item:
                if item['type'] == 'attack-pattern':
                    item_dict[item['id']] = item
        for item in self._RELATIONSHIPS[self.stix]:
            if item in item_dict:
                return_list.append(MobileAttckTechnique(mobile_attck_obj=self.__mobile_attck_obj, **item_dict[item]))
        return return_list

    @property
    def actors(self):
        """
        Accessing actors which have been identified as using a specific tool
        as part of the MITRE Mobile ATT&CK Framework

        Returns:
            list: Returns all actor objects as a list that are associated with a tool
        """
        from .actor import MobileAttckActor
        return_list = []
        item_dict = {}
        for item in self.__mobile_attck_obj['objects']:
            if 'type' in item:
                if item['type'] == 'intrusion-set':
                    item_dict[item['id']] = item
        for item in self._RELATIONSHIPS[self.stix]:
            if item in item_dict:
                return_list.append(MobileAttckActor(mobile_attck_obj=self.__mobile_attck_obj, **item_dict[item]))
        return return_list



from .mobileattckobject import MobileAttckObject


class MobileAttckTechnique(MobileAttckObject):

    """Mobile MITRE ATT&CK Technique object.

    A child class of MobileAttckObject

    Creates objects which have been categorized as a technique used by attackers

    Each technique enables you to access the following properties on the object:

        1. command_list - A list of commands associated with a technique
        2. commands = A list of dictionary objects containing source, command, and
                      provided name associated with a technique
        3. queries = A list of dictionary objects containing product, query, and
                     name associated with a technique
        4. datasets = A list of raw datasets associated with a technique
        5. possible_detections = A list of raw datasets containing possible detection
                                 methods for a technique

    Example:
        You can iterate over an `techniques` list and access specific properties and
        relationship properties.

        The following relationship properties are accessible:
                1. tactics
                2. mitigations
                3. actors

            1. To iterate over an `techniques` list, do the following:

            .. code-block:: python

               from pyattck import Attck

               attck = Attck()

               for technique in attck.mobile.techniques:
                   print(technique.id)
                   print(technique.name)
                   print(technique.alias)
                   print(technique.description)
                   # etc.

            2. To access relationship properties, do the following:

            .. code-block:: python

               from pyattck import Attck

               attck = Attck()

               for technique in attck.mobile.techniques:
                   print(technique.id)
                   print(technique.name)
                   print(technique.alias)
                   print(technique.description)
                   # etc.

                   for malware in technique.malwares:
                       print(malware.name)
                       print(malware.description)
                       # etc.
    """

    def __init__(self, mobile_attck_obj = None, **kwargs):
        """
        Creates an MobileAttckTechnique object.  
        The MobileAttckTechnique object is a technique used by attackers.

        Arguments:
            attck_obj (json) -- Takes the raw MITRE Mobile ATT&CK Json object
            AttckObject (dict) -- Takes the MITRE Mobile ATT&CK Json object as a kwargs values
        """
        super(MobileAttckTechnique, self).__init__(**kwargs)
        self.__mobile_attck_obj = mobile_attck_obj
        self.old_attack_id = self._set_attribute(kwargs, 'x_mitre_old_attack_id')
        self.platforms = self._set_list_items(kwargs, 'x_mitre_platforms')
        self.version = self._set_attribute(kwargs, 'x_mitre_version')
        self.depricated = self._set_attribute(kwargs, 'x_mitre_deprecated')
        self.created_by_ref = self._set_attribute(kwargs, 'created_by_ref')
        self.contributor = self._set_list_items(kwargs, 'x_mitre_contributors')
        self.tactic_type = self._set_list_items(kwargs, 'x_mitre_tactic_type')
        self.external_reference = self._set_reference(kwargs)
        self.possible_detections = self._set_attribute(kwargs, 'x_mitre_detection')
        self.revoked = self._set_attribute(kwargs, 'revoked')
        self.deprecated = self._set_attribute(kwargs, 'x_mitre_deprecated')
        self.wiki = self._set_wiki(kwargs)
        self.stix = self._set_attribute(kwargs, 'id')
        self.command_list = self.__get_filtered_dataset('command_list')
        self.commands = self.__get_filtered_dataset('commands')
        self.queries = self.__get_filtered_dataset('queries')
        self.datasets = self.__get_filtered_dataset('parsed_datasets')
        self.possible_detections = self.__get_filtered_dataset('possible_detections')
        self.tactics = kwargs
        self.set_relationships(self.__mobile_attck_obj)

    def __get_filtered_dataset(self, attribute_name):
        for item in MobileAttckObject.generated_attck_json['techniques']:
            if item['technique_id'] == self.id:
                return item[attribute_name]

    @property
    def tactics(self):
        """
        Accessing tactics that a specific technique belongs to as part
        of the MITRE Mobile ATT&CK Framework

        Returns:
            list: Returns all tactic objects as a list that a technique
                  belongs to
        """
        from .tactic import MobileAttckTactic
        tactic_list = []
        for item in self.__mobile_attck_obj['objects']:
            if 'x-mitre-tactic' in item['type']:
                for tact in self._tactic:
                    if str(tact).lower() == str(item['x_mitre_shortname']).lower():
                        tactic_list.append(MobileAttckTactic(mobile_attck_obj=self.__mobile_attck_obj, **item))
        return tactic_list

    @tactics.setter
    def tactics(self, obj):
        """
        Sets the associated tactic/phase this technique is in

        Arguments:
            obj (dict) -- A Mitre ATT&CK Framework json object

        Returns:
            (string) -- Returns a string that sets the tactic/phase this
                        technique is in. If there is no phase found, it
                        will return 'no phase_name'
        """
        temp_list = []
        try:
            for phase in obj['kill_chain_phases']:
                temp_list.append(phase['phase_name'])
            self._tactic = temp_list
        except:
            self._tactic = ['no phase_name']

    @property
    def mitigations(self):
        """
        Accessing mitigiations for a specific technique as part of the
        MITRE Mobile ATT&CK Framework

        Returns:
            list: Returns all mitigation objects as a list that are associated
                  with a technique
        """
        from .mitigation import MobileAttckMitigation
        return_list = []
        item_dict = {}
        for item in self.__mobile_attck_obj['objects']:
            if 'type' in item:
                if item['type'] == 'course-of-action':
                    item_dict[item['id']] = item
        try:
            for item in self._RELATIONSHIPS[self.stix]:
                if item in item_dict:
                    return_list.append(MobileAttckMitigation(mobile_attck_obj=self.__mobile_attck_obj, **item_dict[item]))
        except:
            pass
        return return_list 

    @property
    def actors(self):
        """
        Accessing actors which have been identified as using a specific
        technique as part of the MITRE Mobile ATT&CK Framework

        Returns:
            list: Returns all actor objects as a list that are associated
                  with a technique
        """
        from .actor import MobileAttckActor
        return_list = []
        item_dict = {}
        for item in self.__mobile_attck_obj['objects']:
            if 'type' in item:
                if item['type'] == 'intrusion-set':
                    item_dict[item['id']] = item
        try:
            for item in self._RELATIONSHIPS[self.stix]:
                if item in item_dict:
                    return_list.append(MobileAttckActor(mobile_attck_obj=self.__mobile_attck_obj, **item_dict[item]))
        except:
            pass
        return return_list



from .attckobject import AttckObject
import attr


@attr.s
class AttckDataComponent:

    created_by_ref = attr.ib()
    modified       = attr.ib()
    created        = attr.ib()
    type           = attr.ib()
    id             = attr.ib()
    name           = attr.ib()
    description    = attr.ib()
    x_mitre_version = attr.ib()
    x_mitre_data_source_ref = attr.ib()
    object_marking_refs = attr.ib()


class AttckDataSource(AttckObject):

    """ICS MITRE ATT&CK Data Source object

    A child class of AttckObject

    Creates objects that are categorized as Mitre ATT&CK ICS
    Data Sources

    Example:
       
    Arguments:
        attck_obj (json) -- Takes the raw Mitre ATT&CK Json object
        AttckObject (dict) -- Takes the Mitre ATT&CK Json object as a
                              kwargs values
    """

    __ATTCK_DATASETS = None

    def __init__(self, attck_obj = None, _data_component_filter=None, _ics_attck_obj=None, **kwargs):
        """
        This class represents a Data Source as defined by the
        ICS MITRE ATT&CK framework.

        Keyword Arguments:
            attck_obj {json} -- A ICS MITRE ATT&CK Framework
                                json object (default: {None})
            _ics_attck_obj {json} -- A ICS MITRE ATT&CK Framework 
                                json object. Used for gathering data_components 
                                and data_sources (default: {None})

        Raises:
            GeneratedDatasetException: Raised an exception when unable
                                       to access or process the external
                                       generated dataset.
        """
        super(AttckDataSource, self).__init__(**kwargs)
        self.__attck_obj = attck_obj
        self.__ics_attck_obj = _ics_attck_obj
        self.__data_component_filter = _data_component_filter
        self.id = self._set_id(kwargs)
        self.created_by_ref = self._set_attribute(kwargs, 'created_by_ref')
        self.name = self._set_attribute(kwargs, 'name')
        self.description = self._set_attribute(kwargs, 'description')
        self.external_reference = self._set_reference(kwargs)
        self.created = self._set_attribute(kwargs, 'created')
        self.modified = self._set_attribute(kwargs, 'modified')
        self.stix = self._set_attribute(kwargs, 'id')
        self.type = self._set_attribute(kwargs, 'type')
        self.wiki = self._set_wiki(kwargs)
        self.contributor = self._set_list_items(kwargs, 'x_mitre_contributors')
        self.platforms = self._set_list_items(kwargs, 'x_mitre_platforms')
        self.collection_layers = self._set_list_items(kwargs, 'x_mitre_collection_layers')
        self.data_components = []
        self.__set_data_component_relationship()

    def __set_data_component_relationship(self):
        if not self.data_components:
            for item in self.__attck_obj['objects']:
                if 'type' in item and item.get('type') == 'x-mitre-data-component':
                    if item['x_mitre_data_source_ref'] == self.stix:
                        if self.__data_component_filter:
                            if item['name'] in self.__data_component_filter:
                                self.data_components.append(AttckDataComponent(**item))
                        else:
                            self.data_components.append(AttckDataComponent(**item))

    @property
    def techniques(self):
        """
        Returns all technique objects as a list that are documented as
        related to a data source

        Returns:
            [list] -- A list of related technique objects defined within the
            ICS MITRE ATT&CK Framework
        """
        from .technique import AttckTechnique
        return_list = []
        for item in self.__ics_attck_obj['objects']:
            if 'type' in item:
                if item['type'] == 'attack-pattern' and item.get('x_mitre_data_sources'):
                    data_sources = self._create_data_sources_dict(item['x_mitre_data_sources'])
                    if data_sources.get(self.name):
                        for component in self.data_components:
                            if component.name in data_sources[self.name]:
                                return_list.append(AttckTechnique(attck_obj=self.__ics_attck_obj, **item))
        return return_list



from .mobileattckobject import MobileAttckObject


class MobileAttckTactic(MobileAttckObject):

    """Mobile MITRE ATT&CK Tactic object.

    A child class of MobileAttckObject

    Creates objects that are categorized as MITRE Mobile ATT&CK Tactics

    Example:

        You can iterate over an `tactics` list and access specific properties and relationship properties.

        The following relationship properties are accessible:
                1. techniques

        1. To iterate over an `tactics` list, do the following:

        .. code-block:: python

            from pyattck import Attck

            attck = Attck()

            for tactic in attck.mobile.tactics:
                print(tactic.id)
                print(tactic.name)
                print(tactic.alias)
                print(tactic.description)
                # etc.

        2. To access relationship properties, do the following:

        .. code-block:: python

            from pyattck import Attck

            attck = Attck()

            for tactic in attck.mobile.tactics:
                print(tactic.id)
                print(tactic.name)
                print(tactic.alias)
                print(tactic.description)
                # etc.

                for technique in tactic.techniques:
                    print(technique.name)
                    print(technique.description)
                    # etc.

    Arguments:
        mobile_attck_obj (json) -- Takes the raw MITRE Mobile ATT&CK Json object
        AttckObject (dict) -- Takes the MITRE Mobile ATT&CK Json object as a kwargs values
    """

    def __init__(self, mobile_attck_obj = None, **kwargs):
        """
        This class represents a Tactic as defined by the
        Mobile MITRE ATT&CK framework.

        Keyword Arguments:
            mobile_attck_obj {json} -- A Mobile MITRE ATT&CK Framework json object (default: {None})
        """
        super(MobileAttckTactic, self).__init__(**kwargs)
        self.__mobile_attck_obj = mobile_attck_obj
        self.created_by_ref = self._set_attribute(kwargs, 'created_by_ref')
        self.short_name = self._set_attribute(kwargs, 'x_mitre_shortname')
        self.external_reference = self._set_reference(kwargs)
        self.stix = self._set_attribute(kwargs, 'id')
        self.wiki = self._set_wiki(kwargs)
        self.set_relationships(self.__mobile_attck_obj)

    @property
    def techniques(self):
        """
        Accessing techniques that belong to a tactic as part of the
        MITRE Mobile ATT&CK Framework

        Returns:
            list: Returns all technique objects as a list that are
                  associated with a tactic object
        """
        from .technique import MobileAttckTechnique
        technique_list = []
        for item in self.__mobile_attck_obj['objects']:
            if 'kill_chain_phases' in item:
                for prop in item['kill_chain_phases']:
                    if str(prop['phase_name']).lower() == str(self.short_name).lower():
                        technique_list.append(MobileAttckTechnique(mobile_attck_obj=self.__mobile_attck_obj, **item))
        return technique_list



class Search:

    def __init__(self, actors, malwares, mitigations, tactics, techniques, tools):
        self.__actors = actors
        self.__malwares = malwares
        self.__mitigations = mitigations
        self.__tactics = tactics
        self.__techniques = techniques
        self.__tools = tools

    def __search(self, kwargs, iterator):
        return_set = set()
        for item in iterator:
            for key,val in kwargs.items():
                if hasattr(item, key) and getattr(item, key):
                    if isinstance(getattr(item, key), list):
                        for i in getattr(item, key):
                            if val in i:
                                return_set.add(item.__repr__())
                    elif isinstance(getattr(item, key), dict):
                        for k,v in getattr(item, key).items():
                            if val in k or val in v:
                                return_set.add(item.__repr__())
                    else:
                        if val in getattr(item, key).lower():
                            return_set.add(item.__repr__())
        return list(return_set)

    def __print_help(self, iterator):
        key_list = []
        for key,val in iterator[0].__dict__.items():
            if not key.startswith('_'):
                key_list.append(key)
        print('Please provide one or more parameters when searching.',
            f'You can provide any property avaible on an object. {key_list}')

    def actors(self, **kwargs):
        if not kwargs:
            self.__print_help(self.__actors)
        return self.__search(kwargs, self.__actors)

    def malwares(self, **kwargs):
        if not kwargs:
            self.__print_help(self.__actors)
        return self.__search(kwargs, self.__malwares)

    def mitigations(self, **kwargs):
        if not kwargs:
            self.__print_help(self.__actors)
        return self.__search(kwargs, self.__mitigations)

    def tactics(self, **kwargs):
        if not kwargs:
            self.__print_help(self.__actors)
        return self.__search(kwargs, self.__tactics)

    def techniques(self, **kwargs):
        if not kwargs:
            self.__print_help(self.__actors)
        return self.__search(kwargs, self.__techniques)

    def tools(self, **kwargs):
        if not kwargs:
            self.__print_help(self.__actors)
        return self.__search(kwargs, self.__tools)



from .technique import AttckTechnique
from .actor import AttckActor
from .malware import AttckMalware
from .tools import AttckTools
from .mitigation import AttckMitigation
from .tactic import AttckTactic
from .control import AttckControl
from .datasource import AttckDataSource
from ..configuration import Configuration



class Enterprise(object):

    """An interface to the Enterprise MITRE ATT&CK Framework.

    This class creates an interface to all data points in the
    MITRE ATT&CK Enterprise framework.

    This interface enables you to retrieve all properties within
    each item in the MITRE ATT&CK Enterprise Framework.

    The following categorical items can be accessed using this class:

        1. Tactics (Tactics are the phases defined by MITRE ATT&CK)
        2. Techniques (Techniques are the individual actions which can
           accomplish a tactic)
        3. Mitigations (Mitigations are recommendations to prevent or
           protect against a technique)
        4. Actors (Actors or Groups are identified malicious actors/groups
           which have been identified and documented by MITRE & third-parties)
        5. Tools (Tools are software used to perform techniques)
        6. Malwares (Malwares are specific pieces of malware used by actors
           (or in general) to accomplish a technique)

    You can also search the external dataset for external commands that are
    similar using the `search_commands` method.

        .. code-block:: python

            from pyattck import Attck

            attck = Attck()

            for search in attck.enterprise.search_commands('powershell'):
                print(search['technique'])
                print(search['reason_for_match'])

    Additionally, as of pyattck 2.0.0 you can now access additional datasets
    related to a technique. These datasets are
    [documented here](https://github.com/swimlane/pyattck/blob/master/generateattcks/README.md).

    Each technique enables you to access the following properties on the object:

        1. command_list - A list of commands associated with a technique
        2. commands = A list of dictionary objects containing source, command,
                      and provided name associated with a technique
        3. queries = A list of dictionary objects containing product, query, and
                     name associated with a technique
        4. datasets = A list of raw datasets associated with a technique
        5. possible_detections = A list of raw datasets containing possible detection
                                 methods for a technique

    Each Actor object (if available) enables you to access the following properties
    on the object:

        1. country
        2. operations
        3. attribution_links
        4. known_tools
        5. targets
        6. additional_comments
        7. external_description

    You can retrieve the entire dataset using the `external_dataset` property on
    a `actor` object.

    pyattck also enables you to retrieve or generate logos for the actor or group
    using the following properties:

        - ascii_logo - Generated ASCII logo based on the actor or groups name
        - image_logo - Generated ASCII logo based on a provided logo

    Each Tools object (if available) enables you to access the following properties
    on the object:

        1. additional_names
        2. attribution_links
        3. additional_comments
        4. family

    You can retrieve the entire dataset using the `external_dataset` property
    on a `tool` object.

    You can also access external data properties from the C2 Matrix project.
    The following properties are generated using C2 Matrix external data:

        - HTTP
        - Implementation
        - Custom Profile
        - DomainFront
        - Multi-User
        - SMB
        - Kill Date
        - macOS
        - GitHub
        - Key Exchange
        - Chaining
        - Price
        - TCP
        - Proxy Aware
        - HTTP3
        - HTTP2
        - Date
        - Evaluator
        - Working Hours
        - Slack
        - FTP
        - Version Reviewed
        - Logging
        - Name
        - License
        - Windows
        - Stego
        - Notes
        - Server
        - Actively Maint.
        - Dashboard
        - DNS
        - Popular Site
        - ICMP
        - IMAP
        - DoH
        - Jitter
        - How-To
        - ATT&CK Mapping
        - Kali
        - Twitter
        - MAPI
        - Site
        - Agent
        - API
        - UI
        - Linux

    You can retrieve the entire dataset using the `c2_data` property.

    As of pyattck 3.0 you can now access defined Compliance Controls related
    to a technique.

    Here is an example of retrieving a list of compliance controls:

        .. code-block:: python

            from pyattck import Attck

            attck = Attck()

            for technique in attck.enterprise.techniques:
                print(technique.id)
                print(technique.name)
                print(technique.description)

                # to get a count of controls for a technique do the following
                print(len(technique.controls))

                # below will print each controls properties & values
                for control in technique.controls:
                    print(control.__dict__)

                # below will print the id, name and description of a control
                for control in technique.controls:
                    print(control.id)
                    print(control.name)
                    print(control.description)

    Example:
        Once an Attck object is instantiated, you can access each object type as
        a list of objects (e.g. techniques, tactics, actors, etc.)

        You can iterate over each object list and access specific properties and
        relationship properties of each.

        The following relationship properties are accessible:
            1. Actors
                1. Tools used by the Actor or Group
                2. Malware used by the Actor or Group
                3. Techniques this Actor or Group uses
            2. Malwares
                1. Actor or Group(s) using this malware
                2. Techniques this malware is used with
            3. Mitigations
                1. Techniques related to a specific set of mitigation suggestions
            4. Tactics
                1. Techniques found in a specific Tactic (phase)
            5. Techniques
                1. Tactics a technique is found in
                2. Mitigation suggestions for a given technique
                3. Actor or Group(s) identified as using this technique
            6. Tools
                1. Techniques that the specified tool is used within
                2. Actor or Group(s) using a specified tool

            1. To iterate over a list, do the following:

            .. code-block:: python

               from pyattck import Attck

               attck = Attck()

               for technique in attck.enterprise.techniques:
                   print(technique.id)
                   print(technique.name)
                   print(technique.description)
                   # etc.

               for mitigation in attck.enterprise.mitigations:
                   print(mitigation.id)
                   print(mitigation.name)
                   print(mitigation.description)
                   # etc.

            2. To access relationship properties, do the following:

            .. code-block:: python

               from pyattck import Attck

               attck = Attck()

               for technique in attck.enterprise.techniques:
                   print(technique.id)
                   print(technique.name)
                   print(technique.description)
                   # etc.

                   for actor in technique.actors:
                       print(actor.id)
                       print(actor.name)
                       print(actor.description)
                       # etc.

               for mitigation in attck.enterprise.mitigations:
                   print(mitigation.id)
                   print(mitigation.name)
                   print(mitigation.description)
                   # etc.

                   for technique in mitigation.techniques:
                       print(technique.name)
                       print(technique.description)
                       # etc.

    Arguments:
        attck_json (json) - The attck_json is supplied by the attck.py module
        when instantiated.

    Returns:
        [Attck]: Returns a Attck object that contains all data from the
                 MITRE ATT&CK Framework
    """

    __tactics = []
    __techniques = []
    __mitigations = []
    __actors = []
    __tools = []
    __malwares = []
    __controls = []
    __data_sources = []
    __ENTERPRISE_GENERATED_DATA_JSON = None
    __nist_controls_json = Configuration.get_data('nist_controls_json')['objects']
    __attck = Configuration.get_data('enterprise_attck_json')


    def __init__(self, nested_subtechniques=True):
        """
        Sets standard properties that are found in all child classes
        as well as provides standard methods used by inherited classes

        Arguments:
            kwargs (dict) -- Takes the MITRE ATT&CK Json object as a kwargs values
            nested_subtechniques (bool) -- Determines if nested subtechniques will
            be used or not. This is passed from attck class
        """
        self.__nested_subtechniques = nested_subtechniques

    @property
    def actors(self):
        """
        Creates AttckActor objects

        Returns:
            (AttckActor) -- (Returns a list of AttckActor objects)
        """
        if not self.__actors:
            for group in self.__attck['objects']:
                if group['type'] == 'intrusion-set':
                    self.__actors.append(AttckActor(attck_obj=self.__attck, **group))
        return self.__actors

    @property
    def controls(self):
        """
        Creates AttckControls objects

        Returns:
            (AttckControl) -- Returns a list of AttckControl objects
        """
        if not self.__controls:
            for control in self.__nist_controls_json:
                if control.get('type') == 'course-of-action':
                    self.__controls.append(AttckControl(attck_obj=self.__attck, **control))
        return self.__controls

    @property
    def data_sources(self):
        """
        Creates AttckDataSource objects

        Returns:
            (AttckDataSource) -- Returns a list of AttckDataSource objects
        """
        if not self.__data_sources:
            for source in self.__attck['objects']:
                if source['type'] == 'x-mitre-data-source':
                    self.__data_sources.append(AttckDataSource(attck_obj=self.__attck, **source))
        return self.__data_sources

    @property
    def tactics(self):
        """
        Creates AttckTactic objects

        Returns:
            (AttckTactic) -- (Returns a list of AttckTactic objects)
        """
        if not self.__tactics:
            for tactic in self.__attck['objects']:
                if tactic['type'] == 'x-mitre-tactic':
                    self.__tactics.append(AttckTactic(attck_obj=self.__attck, **tactic))
        return self.__tactics

    @property
    def mitigations(self):
        """
        Creates AttckMitigation objects

        Returns:
            (AttckMitigation) -- (Returns a list of AttckMitigation objects)
        """
        if not self.__mitigations:
            for mitigation in self.__attck['objects']:
                if mitigation['type'] == 'course-of-action':
                    self.__mitigations.append(AttckMitigation(attck_obj=self.__attck, **mitigation))
        return self.__mitigations

    @property
    def tools(self):
        """
        Creates AttckTools objects

        Returns:
            (AttckTools) -- Returns a list of AttckTools objects
        """
        if not self.__tools:
            for tools in self.__attck['objects']:
                if tools['type'] == 'tool':
                    self.__tools.append(AttckTools(attck_obj=self.__attck, **tools))
        return self.__tools

    @property
    def malwares(self):
        """
        Creates AttckMalware objects

        Returns:
            (AttckMalware) -- Returns a list of AttckMalware objects
        """
        if not self.__malwares:
            for malware in self.__attck['objects']:
                if malware['type'] == 'malware':
                    self.__malwares.append(AttckMalware(attck_obj=self.__attck, **malware))
        return self.__malwares

    @property
    def techniques(self):
        """
        Creates AttckTechnique objects

        Returns:
            (AttckTechnique) -- Returns a list of AttckTechnique objects
        """
        if not self.__techniques:
            subtechniques = []
            for technique in self.__attck["objects"]:
                if technique.get('type') == 'attack-pattern' and technique.get('revoked') is not True:
                    if self.__nested_subtechniques:
                        if technique.get('x_mitre_is_subtechnique'):
                            subtechniques.append(technique)
                        else:
                            self.__techniques.append(AttckTechnique(attck_obj=self.__attck, **technique))
                    else:
                        self.__techniques.append(AttckTechnique(attck_obj=self.__attck, **technique))
            if subtechniques:
                for item in subtechniques:
                    if item.get('external_references'):
                        for p in item.get('external_references'):
                            for s in p:
                                if p[s] == 'mitre-attack':
                                    for technique in self.__techniques:
                                        if p['external_id'].split('.')[0] == technique.id:
                                            technique.subtechniques = AttckTechnique(attck_obj=self.__attck, **item)
        return self.__techniques

    @property
    def search(self):
        from .search import Search
        return Search(
            actors=self.actors,
            malwares=self.malwares,
            mitigations=self.mitigations,
            tactics=self.tactics,
            techniques=self.techniques,
            tools=self.tools
        )

    def search_commands(self, search_term, json=False):
        """
        Search external datasets for potential commands using a search term

        Args:
            search_term (str): A command to search for close matches against
                               all external datasets containing potential commands

        Returns:
            list: A list of dictionaries containing the technique and the reason
                  for a close match
        """
        if json:
            import json
        return_list = []
        if not self.__ENTERPRISE_GENERATED_DATA_JSON:
            self.__ENTERPRISE_GENERATED_DATA_JSON = Configuration.get_data('generated_attck_json')
        for item in self.__ENTERPRISE_GENERATED_DATA_JSON['techniques']:
            if 'command_list' in item:
                if item['command_list']:
                    for cmd in item['command_list']:
                        if cmd:
                            if search_term in cmd:
                                for technique in self.techniques:
                                    if technique.id.lower() == item['technique_id'].lower():
                                        if json:
                                            return_list.append({
                                                'technique': json.dumps(str(technique))
                                            })
                                        else:
                                            return_list.append({
                                                'technique': technique,
                                                'command_list': technique.command_list
                                            })
        if return_list:
            return return_list
        else:
            return 'No similar commands found'



class GeneratedDatasetException(Exception):

    """
    Raised when unable to retrieve generated datasets and related properties
    """
    pass

class ConfigurationException(Exception):

    """Raised when unable to load or read the pyattck configuration file.
    """
    pass

class UknownFileError(ValueError):
    """Raised when the provided file extension is unkown or is not json, yml or yaml
    """
    def __init__(self, provided_value=None, known_values=None):
        if provided_value and known_values:
            if isinstance(known_values, list):
                super().__init__("The provided value {} is unknown. Please provide a file path with one of these '{}' extensions.".format(
                    provided_value,
                    ','.join([x for x in known_values])
                ))
        else:
            pass


from .preattckobject import PreAttckObject


class PreAttckTactic(PreAttckObject):

    """MITRE PRE-ATT&CK Tactic object.

    A child class of PreAttckObject

    Creates objects that are categorized as MITRE PRE-ATT&CK Tactics

    Example:

        You can iterate over an `tactics` list and access specific properties
        and relationship properties.

        The following relationship properties are accessible:
                1. techniques

        1. To iterate over an `tactics` list, do the following:

        .. code-block:: python

            from pyattck import Attck

            attck = Attck()

            for tactic in attck.preattack.tactics:
                print(tactic.id)
                print(tactic.name)
                print(tactic.alias)
                print(tactic.description)
                # etc.

        2. To access relationship properties, do the following:

        .. code-block:: python

            from pyattck import Attck

            attck = Attck()

            for tactic in attck.preattack.tactics:
                print(tactic.id)
                print(tactic.name)
                print(tactic.alias)
                print(tactic.description)
                # etc.

                for technique in tactic.techniques:
                    print(technique.name)
                    print(technique.description)
                    # etc.

    Arguments:
        attck_obj (json) -- Takes the raw MITRE PRE-ATT&CK Json object
        AttckObject (dict) -- Takes the MITRE PRE-ATT&CK Json object as a kwargs values
    """

    def __init__(self, preattck_obj = None, **kwargs):
        """
        This class represents a Tactic as defined by the
        MITRE PRE-ATT&CK framework.

        Keyword Arguments:
            preattck_obj {json} -- A MITRE PRE-ATT&CK Framework
                                   json object (default: {None})
        """
        super(PreAttckTactic, self).__init__(**kwargs)
        self.__preattck_obj = preattck_obj
        self.created_by_ref = self._set_attribute(kwargs, 'created_by_ref')
        self.stix = self._set_attribute(kwargs, 'id')
        self.short_name = self._set_attribute(kwargs, 'x_mitre_shortname')
        self.wiki = self._set_wiki(kwargs)
        self.set_relationships(self.__preattck_obj)

    @property
    def techniques(self):
        """
        Accessing techniques a specific tactic belongs to based
        on the MITRE PRE-ATT&CK Framework

        Returns:
            list: Returns all technique objects within the defined tactic
        """
        from .technique import PreAttckTechnique
        technique_list = []
        for item in self.__preattck_obj['objects']:
            if 'kill_chain_phases' in item:
                for prop in item['kill_chain_phases']:
                    if str(prop['phase_name']).lower() == str(self.short_name).lower():
                        technique_list.append(PreAttckTechnique(preattck_obj=self.__preattck_obj, **item))
        return technique_list



from ..configuration import Configuration


class AttckObject(object):

    """Parent class of all other Mitre ATT&CK based classes

    This is a private class and should not be accessed directly
    
    Arguments:
        AttckObject (dict) -- Takes the Mitre ATT&CK Json object as
                              a kwargs values
    """

    _RELATIONSHIPS = {}
    generated_attck_json = Configuration.get_data('generated_attck_json')
    nist_controls_json = Configuration.get_data('nist_controls_json')['objects']
    generated_nist_json = Configuration.get_data('generated_nist_json')

    def __init__(self, **kwargs):
        """
        Sets standard properties that are found in all child classes as well as
        provides standard methods used by inherited classes

        Arguments:
            kwargs (dict) -- Takes the Mitre ATT&CK Json object as a kwargs values
        """
        self.id = self._set_id(kwargs)
        self.name = self._set_attribute(kwargs, 'name')
        self.alias = self.__set_alias(kwargs)
        self.aliases = self.alias
        self.description = self._set_attribute(kwargs, 'description')
        self.reference = self._set_reference(kwargs)
        self.created = self._set_attribute(kwargs, 'created')
        self.modified = self._set_attribute(kwargs, 'modified')
        self.stix = self._set_attribute(kwargs, 'id')
        self.type = self._set_attribute(kwargs, 'type')

    def __str__(self):
        """
        Returns dictionary string of all properties and
        values for the instance

        Returns:
            (str): All properties and values of instance
        """
        return_dict = {}
        for key,val in self.__dict__.items():
            if not key.startswith('_'):
                return_dict[key] = val
        return str(return_dict)

    def __repr__(self):
        """
        Returns a printable representation of an object

        Returns:
            str: Returns a printable representation of an object
        """
        return "{class_name}('{name}', '{id}')".format(
            class_name=self.__class__.__name__,
            name=self.name,
            id=self.id
        )

    def set_relationships(self, attck_obj, update=False):
        """
        Generates relationships within attck_obj based on a defined relationship
        from Mitre ATT&CK

        Args:
            attck_obj (dict): MITRE ATT&CK Json object
        """
        if not AttckObject._RELATIONSHIPS or update:
            for item in attck_obj['objects']:
                if 'type' in item:
                    if item['type'] == 'relationship':
                        source_id = item['source_ref']
                        target_id = item['target_ref']
                        if source_id not in AttckObject._RELATIONSHIPS:
                            AttckObject._RELATIONSHIPS[source_id] = []
                        AttckObject._RELATIONSHIPS[source_id].append(target_id)
                        if target_id not in AttckObject._RELATIONSHIPS:
                            AttckObject._RELATIONSHIPS[target_id] = []
                        AttckObject._RELATIONSHIPS[target_id].append(source_id)

    def _create_data_sources_dict(self, obj):
        return_dict = {}
        if isinstance(obj, list):
            for item in obj:
                if ':' in item:
                    data_source, data_component = item.split(':')
                    if data_source not in return_dict:
                        return_dict[data_source] = []
                    return_dict[data_source].append(data_component.strip())
        return return_dict

    def __set_alias(self, obj):
        """
        Returns the Mitre ATT&CK Framework aliases

        Arguments:
            obj (dict) -- A Mitre ATT&CK Framework json object

        Returns:
            (str) -- Returns the Mitre ATT&CK Framework aliases
        """
        return_list = []
        if obj.get('aliases'):
            for item in obj['aliases']:
                return_list.append(item)
        if obj.get('x_mitre_aliases'):
            for item in obj['x_mitre_aliases']:
                return_list.append(item)
        return return_list

    def _set_attribute(self, obj, name):
        """
        Parent class method to set attribute based on passed in object
           and the name of the property

        Arguments:
            obj (dict) -- Provided json objects are passed to this method
            name (str) -- The json property name to set attribute in child classes

        Returns:
            (str) -- Returns either the value of the attribute requested or
                     returns 'null'
        """
        try:
            value = obj.get(name)
            return None if not value else value
        except:
            return None

    def _set_list_items(self, obj, list_name):
        """
        Private method used by child classes and normalizes list items

        Args:
            obj (dict) -- Provided json objects are passed to this method
            list_name (str) -- The json property name to set list items
                               attribute in child classes

        Returns:
            list: returns a list of values from the provided list_name property
        """
        item_value = []
        if list_name in obj:
            for item in obj[list_name]:
                item_value.append(item)
            return item_value

    def _set_id(self, obj):
        """
        Returns the Mitre ATT&CK Framework external ID

        Arguments:
            obj (dict) -- A Mitre ATT&CK Framework json object

        Returns:
            (str) -- Returns the Mitre ATT&CK Framework external ID
        """
        if obj.get('external_references'):
            for p in obj['external_references']:
                if p.get('source_name') == 'mitre-ics-attack':
                    return p.get('external_id')
                elif 'NIST' in p.get('source_name'):
                    return p.get('external_id')
        return 'No ID Defined'

    def _set_wiki(self, obj):
        """
        Returns the Mitre ATT&CK Framework Wiki URL

        Arguments:
            obj (dict) -- A Mitre ATT&CK Framework json object

        Returns:
            (str) -- Returns the Mitre ATT&CK Framework Wiki URL
        """
        if obj.get('external_references'):
            for p in obj['external_references']:
                if p.get('source_name') == 'mitre-attack':
                    return p.get('url')

    def _set_reference(self, obj):
        """
        Returns a list of external references from the provided
        Mitre ATT&CK Framework json object

        Arguments:
            obj (dict) -- A Mitre ATT&CK Framework json object

        Returns:
            (dict) -- Returns a dict containing the following key/value pairs
                external_id (str) -- The Mitre ATT&CK Framework external ID
                url (str)         -- The Mitre ATT&CK Framework URL
                source_name (str) -- The Mitre ATT&CK Framework source name
                description (str) -- The Mitre ATT&CK Framework description
                                     or None if it does not exist
        """
        return_list = []
        if obj.get('external_references'):
            for p in obj['external_references']:
                return_list.append(p)
        return return_list



from .technique import AttckTechnique
from .malware import AttckMalware
from .mitigation import AttckMitigation
from .tactic import AttckTactic
from .control import AttckControl
from .datasource import AttckDataSource
from ..configuration import Configuration


class ICSAttck(object):

    """An interface to the ICS MITRE ATT&CK Framework.

    This class creates an interface to all data points in the
    MITRE ATT&CK ICS framework.

    This interface enables you to retrieve all properties within
    each item in the MITRE ATT&CK ICS Framework.

    The following categorical items can be accessed using this class:

        1. Tactics (Tactics are the phases defined by MITRE ATT&CK)
        2. Techniques (Techniques are the individual actions which can
           accomplish a tactic)
        3. Mitigations (Mitigations are recommendations to prevent or
           protect against a technique)
        4. Malwares (Malwares are specific pieces of malware used by actors
           (or in general) to accomplish a technique)

    You can also search the external dataset for external commands that are
    similar using the `search_commands` method.

        .. code-block:: python

            from pyattck import Attck

            attck = Attck()

            for search in attck.enterprise.search_commands('powershell'):
                print(search['technique'])
                print(search['reason_for_match'])

    Additionally, as of pyattck 2.0.0 you can now access additional datasets
    related to a technique. These datasets are
    [documented here](https://github.com/swimlane/pyattck/blob/master/generateattcks/README.md).

    Each technique enables you to access the following properties on the object:

        1. command_list - A list of commands associated with a technique
        2. commands = A list of dictionary objects containing source, command,
                      and provided name associated with a technique
        3. queries = A list of dictionary objects containing product, query, and
                     name associated with a technique
        4. datasets = A list of raw datasets associated with a technique
        5. possible_detections = A list of raw datasets containing possible detection
                                 methods for a technique

    As of pyattck 3.0 you can now access defined Compliance Controls related
    to a technique.

    Here is an example of retrieving a list of compliance controls:

        .. code-block:: python

            from pyattck import Attck

            attck = Attck()

            for technique in attck.ics.techniques:
                print(technique.id)
                print(technique.name)
                print(technique.description)

                # to get a count of controls for a technique do the following
                print(len(technique.controls))

                # below will print each controls properties & values
                for control in technique.controls:
                    print(control.__dict__)

                # below will print the id, name and description of a control
                for control in technique.controls:
                    print(control.id)
                    print(control.name)
                    print(control.description)

    Example:
        Once an Attck object is instantiated, you can access each object type as
        a list of objects (e.g. techniques, tactics, actors, etc.)

        You can iterate over each object list and access specific properties and
        relationship properties of each.

        The following relationship properties are accessible:
            1. Malwares
                1. Actor or Group(s) using this malware
                2. Techniques this malware is used with
            2. Mitigations
                1. Techniques related to a specific set of mitigation suggestions
            3. Tactics
                1. Techniques found in a specific Tactic (phase)
            4. Techniques
                1. Tactics a technique is found in
                2. Mitigation suggestions for a given technique
                3. Actor or Group(s) identified as using this technique

            1. To iterate over a list, do the following:

            .. code-block:: python

               from pyattck import Attck

               attck = Attck()

               for technique in attck.ics.techniques:
                   print(technique.id)
                   print(technique.name)
                   print(technique.description)
                   # etc.

               for mitigation in attck.ics.mitigations:
                   print(mitigation.id)
                   print(mitigation.name)
                   print(mitigation.description)
                   # etc.

            2. To access relationship properties, do the following:

            .. code-block:: python

               from pyattck import Attck

               attck = Attck()

               for technique in attck.ics.techniques:
                   print(technique.id)
                   print(technique.name)
                   print(technique.description)
                   # etc.

                   for actor in technique.actors:
                       print(actor.id)
                       print(actor.name)
                       print(actor.description)
                       # etc.

               for mitigation in attck.ics.mitigations:
                   print(mitigation.id)
                   print(mitigation.name)
                   print(mitigation.description)
                   # etc.

                   for technique in mitigation.techniques:
                       print(technique.name)
                       print(technique.description)
                       # etc.

    Arguments:
        attck_json (json) - The attck_json is supplied by the attck.py module
        when instantiated.

    Returns:
        [Attck]: Returns a Attck object that contains all data from the
                 MITRE ATT&CK Framework
    """

    __tactics = []
    __techniques = []
    __mitigations = []
    __malwares = []
    __controls = []
    __data_sources = []
    __ENTERPRISE_GENERATED_DATA_JSON = None
    __nist_controls_json = Configuration.get_data('nist_controls_json')['objects']
    __attck = Configuration.get_data('ics_attck_json')
    __enterprise_attck = Configuration.get_data('enterprise_attck_json')

    @property
    def controls(self):
        """
        Creates AttckControls objects

        Returns:
            (AttckControl) -- Returns a list of AttckControl objects
        """
        if not self.__controls:
            for control in self.__nist_controls_json:
                if control.get('type') == 'course-of-action':
                    self.__controls.append(AttckControl(attck_obj=self.__attck, **control))
        return self.__controls

    @property
    def data_sources(self):
        """
        Creates AttckDataSource objects

        Returns:
            (AttckDataSource) -- Returns a list of AttckDataSource objects
        """
        if not self.__data_sources:
            for source in self.__enterprise_attck['objects']:
                if source['type'] == 'x-mitre-data-source':
                    self.__data_sources.append(AttckDataSource(attck_obj=self.__enterprise_attck, _ics_attck_obj=self.__attck, **source))
        return self.__data_sources

    @property
    def tactics(self):
        """
        Creates AttckTactic objects

        Returns:
            (AttckTactic) -- (Returns a list of AttckTactic objects)
        """
        if not self.__tactics:
            for tactic in self.__attck['objects']:
                if tactic['type'] == 'x-mitre-tactic':
                    self.__tactics.append(AttckTactic(attck_obj=self.__attck, **tactic))
        return self.__tactics

    @property
    def mitigations(self):
        """
        Creates AttckMitigation objects

        Returns:
            (AttckMitigation) -- (Returns a list of AttckMitigation objects)
        """
        if not self.__mitigations:
            for mitigation in self.__attck['objects']:
                if mitigation['type'] == 'course-of-action':
                    self.__mitigations.append(AttckMitigation(attck_obj=self.__attck, **mitigation))
        return self.__mitigations

    @property
    def malwares(self):
        """
        Creates AttckMalware objects

        Returns:
            (AttckMalware) -- Returns a list of AttckMalware objects
        """
        if not self.__malwares:
            for malware in self.__attck['objects']:
                if malware['type'] == 'malware':
                    self.__malwares.append(AttckMalware(attck_obj=self.__attck, **malware))
        return self.__malwares

    @property
    def techniques(self):
        """
        Creates AttckTechnique objects

        Returns:
            (AttckTechnique) -- Returns a list of AttckTechnique objects
        """
        if not self.__techniques:
            for technique in self.__attck["objects"]:
                if technique.get('type') == 'attack-pattern':
                    self.__techniques.append(AttckTechnique(attck_obj=self.__attck, _enterprise_attck_obj=self.__enterprise_attck, **technique))
        return self.__techniques

    def search_commands(self, search_term, json=False):
        """
        Search external datasets for potential commands using a search term

        Args:
            search_term (str): A command to search for close matches against
                               all external datasets containing potential commands

        Returns:
            list: A list of dictionaries containing the technique and the reason
                  for a close match
        """
        if json:
            import json
        return_list = []
        if not self.__ENTERPRISE_GENERATED_DATA_JSON:
            self.__ENTERPRISE_GENERATED_DATA_JSON = Configuration.get_data('generated_attck_json')
        for item in self.__ENTERPRISE_GENERATED_DATA_JSON['techniques']:
            if 'command_list' in item:
                if item['command_list']:
                    for cmd in item['command_list']:
                        if cmd:
                            if search_term in cmd:
                                for technique in self.techniques:
                                    if technique.id.lower() == item['technique_id'].lower():
                                        if json:
                                            return_list.append({
                                                'technique': json.dumps(str(technique))
                                            })
                                        else:
                                            return_list.append({
                                                'technique': technique,
                                                'command_list': technique.command_list
                                            })
        if return_list:
            return return_list
        else:
            return 'No similar commands found'



from .technique import PreAttckTechnique
from .actor import PreAttckActor
from .tactic import PreAttckTactic
from ..configuration import Configuration


class PreAttck(object):

    """MITRE PRE-ATT&CK interface.

    This class creates an interface to all data points in the
    MITRE PRE-ATT&CK framework.

    This interface enables you to retrieve all properties within
    each item in the MITRE PRE-ATT&CK Framework.

    The following categorical items can be accessed using this class:

        1. Tactics (Tactics are the phases defined by MITRE ATT&CK)
        2. Techniques (Techniques are the individual actions which can
                       accomplish a tactic)
        3. Actors (Actors or Groups are identified malicious actors/groups
                   which have been identified and documented by MITRE & third-parties)

    Each Actor object (if available) enables you to access the following properties
    on the object:

        1. country
        2. operations
        3. attribution_links
        4. known_tools
        5. targets
        6. additional_comments
        7. external_description

    You can retrieve the entire dataset using the `external_dataset` property
    on a `actor` object.

    pyattck also enables you to retrieve or generate logos for the actor or group
    using the following properties:

        - ascii_logo - Generated ASCII logo based on the actor or groups name

    Example:
        Once an Attck object is instantiated, you can access each object type
        as a list of objects (e.g. techniques, tactics, actors, etc.)

        You can iterate over each object list and access specific properties
        and relationship properties of each.

        The following relationship properties are accessible:
            1. Actors
                1. Techniques this Actor or Group uses
            2. Tactics
                1. Techniques found in a specific Tactic (phase)
            3. Techniques
                1. Tactics a technique is found in
                2. Actor or Group(s) identified as using this technique

            1. To iterate over a list, do the following:

            .. code-block:: python

               from pyattck import Attck

               attck = Attck()

               for technique in attck.preattack.techniques:
                   print(technique.id)
                   print(technique.name)
                   print(technique.description)
                   # etc.

            2. To access relationship properties, do the following:

            .. code-block:: python

               from pyattck import Attck

               attck = Attck()

               for technique in attck.preattack.techniques:
                   print(technique.id)
                   print(technique.name)
                   print(technique.description)
                   # etc.

                   for actor in technique.actors:
                       print(actor.id)
                       print(actor.name)
                       print(actor.description)
                       # etc.
    """

    __tactics = []
    __techniques = []
    __actors = []
    __preattck_json = Configuration.get_data('pre_attck_json')

    @property
    def actors(self):
        """
        Access all actors within the MITRE PRE-ATT&CK Framework

        Returns:
            PreAttckActor: Returns a list of PreAttckActor objects
        """
        if not self.__actors:
            for group in self.__preattck_json['objects']:
                if group['type'] == 'intrusion-set':
                    self.__actors.append(PreAttckActor(preattck_obj=self.__preattck_json, **group))
        return self.__actors

    @property
    def tactics(self):
        """
        Access all tactics within the MITRE PRE-ATT&CK Framework

        Returns:
            PreAttckTactic: Returns a list of PreAttckTactic objects
        """
        if not self.__tactics:
            for tactic in self.__preattck_json['objects']:
                if tactic['type'] == 'x-mitre-tactic':
                    self.__tactics.append(PreAttckTactic(preattck_obj=self.__preattck_json, **tactic))
        return self.__tactics

    @property
    def techniques(self):
        """
        Access all techniques within the MITRE PRE-ATT&CK Framework

        Returns:
            PreAttckTechnique: Returns a list of PreAttckTechnique objects
        """
        if not self.__techniques:
            for technique in self.__preattck_json["objects"]:
                if (technique['type'] == 'attack-pattern'):
                    self.__techniques.append(PreAttckTechnique(preattck_obj=self.__preattck_json, **technique))
        return self.__techniques



from .attckobject import AttckObject


class AttckControl(AttckObject):

    """An object that represents a compliance control type.

    Currently this object is only utilized by NIST 800-53 controls
    but will be expanded in the future.

    Each control relates to a technique and can assist with identifying
    compliance needs and coverage for specific techniques
    and sub-techniques.

    Example:
        You can iterate over an `techniques` list and access specific
        control properties.

            1. To iterate over an `techniques` list, do the following:

            .. code-block:: python

               from pyattck import Attck

               attck = Attck()

               for technique in attck.enterprise.techniques:
                   print(technique.id)
                   print(technique.name)
                   print(technique.alias)
                   print(technique.description)
                   # etc.
                   # you can access controls related to a technique
                   for control in technique.controls:
                       print(control.id)
                       print(control.name)
                       print(control.description)
                       # etc.

    Arguments:
        kwargs (dict) -- Takes the raw control Json object
    """

    def __init__(self, attck_obj = None, **kwargs):
        """
        This class represents a compliance Control as defined from
        external data sources.

        Keyword Arguments:
            kwargs (dict) -- A compliance control JSON object
        """
        super(AttckControl, self).__init__(**kwargs)
        self.__attck_obj = attck_obj
        for key,val in kwargs.items():
            prop_name = key.replace('x_mitre_','')
            if not hasattr(self, prop_name):
                setattr(self, prop_name, val)

    @property
    def techniques(self):
        """
        Returns all technique objects as a list that are
        associated with a Tactic

        Returns:
            [list] -- A list of related technique objects defined
                      within the Enterprise MITRE ATT&CK Framework
        """
        from .technique import AttckTechnique
        technique_list = []
        for key,val in AttckObject.generated_nist_json.items():
            if self.stix in val:
                for item in self.__attck_obj['objects']:
                    if 'type' in item and item['type'] == 'attack-pattern' and key == item['id']:
                        technique_list.append(AttckTechnique(attck_obj=self.__attck_obj, **item))
        return technique_list



from .mobileattckobject import MobileAttckObject


class MobileAttckMalware(MobileAttckObject):

    """Mobile MITRE ATT&CK Malware object.

    A child class of MobileAttckObject

    Creates objects which have been categorized as malware used in attacks

    Example:
        You can iterate over a `malwares` list and access specific properties
        and relationship properties.

        The following relationship properties are accessible:
                1. actors
                2. techniques

            1. To iterate over an `malwares` list, do the following:

            .. code-block:: python

               from pyattck import Attck

               attck = Attck()

               for malware in attck.mobile.malwares:
                   print(malware.id)
                   print(malware.name)
                   print(malware.alias)
                   print(malware.description)
                   # etc.

            2. To access relationship properties, do the following:

            .. code-block:: python

               from pyattck import Attck

               attck = Attck()

               for malware in attck.mobile.malwares:
                   print(malware.id)
                   print(malware.name)
                   print(malware.alias)
                   print(malware.description)

                   for technique in malware.techniques:
                       print(technique.name)
                       print(technique.description)
                       # etc.
    """

    def __init__(self, mobile_attck_obj = None, **kwargs):
        """
        Creates an MobileAttckMalware object.  
        The MobileAttckMalware object is based on malware which have been categorized
        as software used in attacks

        Arguments:
            mobile_attck_obj (json) -- Takes the raw MITRE Mobile ATT&CK Json object
            AttckObject (dict) -- Takes the MITRE Mobile ATT&CK Json object as a kwargs values
        """
        super(MobileAttckMalware, self).__init__(**kwargs)
        self.__mobile_attck_obj = mobile_attck_obj
        self.old_attack_id = self._set_attribute(kwargs, 'x_mitre_old_attack_id')
        self.platforms = self._set_list_items(kwargs, 'x_mitre_platforms')
        self.version = self._set_attribute(kwargs, 'x_mitre_version')
        self.created_by_ref = self._set_attribute(kwargs, 'created_by_ref')
        self.labels = self._set_list_items(kwargs, 'labels')
        self.external_references = self._set_reference(kwargs)
        self.contributor = self._set_list_items(kwargs, 'x_mitre_contributors')
        self.stix = self._set_attribute(kwargs, 'id')
        self.wiki = self._set_wiki(kwargs)
        self.set_relationships(self.__mobile_attck_obj)

    @property
    def actors(self):
        """
        Accessing known actors who have used a piece of malware as part of the
        MITRE Mobile ATT&CK Framework

        Returns:
            list: Returns all actor objects as a list that are documented using a
                  piece of malware
        """
        from .actor import MobileAttckActor
        return_list = []
        item_dict = {}
        for item in self.__mobile_attck_obj['objects']:
            if 'type' in item:
                if item['type'] == 'intrusion-set':
                    item_dict[item['id']] = item
        for item in self._RELATIONSHIPS[self.stix]:
            if item in item_dict:
                return_list.append(MobileAttckActor(mobile_attck_obj=self.__mobile_attck_obj, **item_dict[item]))
        return return_list

    @property
    def techniques(self):
        """Accessing known techniques a piece of malware has been used in as part of the MITRE Mobile ATT&CK Framework

        Returns:
            list: Returns all technique objects as a list that are documented as using a piece of malware
        """
        from .technique import MobileAttckTechnique
        return_list = []
        item_dict = {}
        for item in self.__mobile_attck_obj['objects']:
            if 'type' in item:
                if item['type'] == 'attack-pattern':
                    item_dict[item['id']] = item
        for item in self._RELATIONSHIPS[self.stix]:
            if item in item_dict:
                return_list.append(MobileAttckTechnique(mobile_attck_obj=self.__mobile_attck_obj, **item_dict[item]))
        return return_list



import os
import json
import warnings
from urllib.parse import urlparse
from pathlib import Path
import yaml
from requests.api import request
from .utils.exceptions import UknownFileError


class ConfigurationProperties(type):

    __config_data = None

    def __download_url_data(cls, url):
        response = request('GET', url, **cls.requests_kwargs)
        if response.status_code == 200:
            return response.json()
        return {}

    def _check_if_path(cls, value):
        if Path(value):
            return True
        return False

    def _check_if_url(cls, value):
        try:
            if urlparse(value).scheme in ['http', 'https']:
                return True
            return False
        except:
            return False

    def __get_absolute_path(cls, path_string):
        return os.path.abspath(
            os.path.expanduser(
                os.path.expandvars(path_string)
            )
        )

    def __validate_value_string(cls, value):
        if cls._check_if_url(value):
            return value
        elif cls._check_if_path(value):
            return value
        else:
            raise Exception('The provided value is neither a URL or file path')

    def __write_to_disk(cls, path, data):
        if not os.path.exists(os.path.dirname(path)):
            try:
                os.makedirs(os.path.dirname(path))
            except:
                raise Exception('pyattck attempted to create the provided directories but was unable to: {}'.format(path))
        with open(path, 'w+') as f:
            if path.endswith('.json'):
                json.dump(data, f)
            elif path.endswith('.yml') or path.endswith('.yaml'):
                yaml.dump(data, f)
            else:
                raise UknownFileError(provided_value=path, known_values=['.json', '.yml', '.yaml'])

    def __read_from_disk(cls, path):
        if os.path.exists(path) and os.path.isfile(path):
            try:
                with open(path) as f:
                    if path.endswith('.json'):
                        return json.load(f)
                    elif path.endswith('.yml') or path.endswith('.yaml'):
                        return yaml.load(f, Loader=yaml.FullLoader)
                    else:
                        raise UknownFileError(provided_value=path, known_values=['.json', '.yml', '.yaml'])
            except:
                pass
        return None

    def _save_json_data(cls, force: bool=False) -> None:
        if not os.path.exists(cls.data_path):
            try:
                os.makedirs(cls.data_path)
            except:
                raise Exception(
                    'Unable to save data to the provided location: {}'.format(cls.data_path)
                )
        for json_data in ['enterprise_attck_json', 'pre_attck_json', 
                          'mobile_attck_json', 'ics_attck_json', 'nist_controls_json', 
                          'generated_attck_json', 'generated_nist_json']:
            if cls._check_if_url(getattr(cls, json_data)):
                try:
                    path = os.path.join(cls.data_path, "{json_data}.json".format(json_data=json_data))
                    if not os.path.exists(path) or force:
                        data = cls.__download_url_data(getattr(cls, json_data))
                        cls.__write_to_disk(path, data)
                    setattr(cls, '_' + json_data, path)
                except:
                    raise Warning(f"Unable to download data from {json_data}")
        cls.__update_config()

    def __update_config(cls):
        cls.__config_data = {
            'data_path': cls.data_path,
            'enterprise_attck_json': cls._enterprise_attck_json,
            'pre_attck_json': cls._pre_attck_json,
            'mobile_attck_json': cls._mobile_attck_json,
            'ics_attck_json': cls._ics_attck_json,
            'nist_controls_json': cls._nist_controls_json,
            'generated_attck_json': cls._generated_attck_json,
            'generated_nist_json': cls._generated_nist_json,
            'config_file_path': cls._config_file_path
        }

    def get_data(cls, value: str) -> dict:
        if cls._check_if_url(cls.config_data.get(value)):
            return cls.__download_url_data(cls.config_data.get(value))
        else:
            return cls.__read_from_disk(cls.config_data.get(value))

    @property
    def requests_kwargs(cls):
        return cls._requests_kwargs

    @requests_kwargs.setter
    def requests_kwargs(cls, value):
        cls._requests_kwargs = value

    @property
    def use_config(cls):
        return cls._use_config

    @use_config.setter
    def use_config(cls, value):
        cls._use_config = bool(value)

    @property
    def save_config(cls):
        return cls._save_config

    @save_config.setter
    def save_config(cls, value):
        cls._save_config = bool(value)

    @property
    def config_file_path(cls):
        return cls.__get_absolute_path(cls._config_file_path)

    @config_file_path.setter
    def config_file_path(cls, value):
        cls._config_file_path = cls.__get_absolute_path(value)
        cls.__update_config()

    @property
    def data_path(cls):
        return cls.__get_absolute_path(cls._data_path)

    @data_path.setter
    def data_path(cls, value):
        cls._data_path = cls.__get_absolute_path(value)
        cls.__update_config()

    @property
    def config_data(cls):
        if cls.use_config:
            cls.__config_data = cls.__read_from_disk(cls.config_file_path)
            if not cls.__config_data:
                cls.__update_config()
        else:
            cls.__update_config()
        if cls.save_config:
            cls.__update_config()
            cls.__write_to_disk(cls.config_file_path, cls.__config_data)
            cls._save_json_data()
        return cls.__config_data

    @property
    def enterprise_attck_json(cls):
        return cls._enterprise_attck_json

    @enterprise_attck_json.setter
    def enterprise_attck_json(cls, value):
        cls._enterprise_attck_json = cls.__validate_value_string(value)
        cls.__update_config()

    @property
    def pre_attck_json(cls):
        return cls._pre_attck_json

    @pre_attck_json.setter
    def pre_attck_json(cls, value):
        cls._pre_attck_json = cls.__validate_value_string(value)
        cls.__update_config()

    @property
    def mobile_attck_json(cls):
        return cls._mobile_attck_json

    @mobile_attck_json.setter
    def mobile_attck_json(cls, value):
        cls._mobile_attck_json = cls.__validate_value_string(value)
        cls.__update_config()

    @property
    def ics_attck_json(cls):
        return cls._ics_attck_json

    @ics_attck_json.setter
    def ics_attck_json(cls, value):
        cls._ics_attck_json = cls.__validate_value_string(value)
        cls.__update_config()

    @property
    def nist_controls_json(cls):
        return cls._nist_controls_json

    @nist_controls_json.setter
    def nist_controls_json(cls, value):
        cls._nist_controls_json = cls.__validate_value_string(value)
        cls.__update_config()

    @property
    def generated_attck_json(cls):
        return cls._generated_attck_json

    @generated_attck_json.setter
    def generated_attck_json(cls, value):
        cls._generated_attck_json = cls.__validate_value_string(value)
        cls.__update_config()

    @property
    def generated_nist_json(cls):
        return cls._generated_nist_json

    @generated_nist_json.setter
    def generated_nist_json(cls, value):
        cls._generated_nist_json = cls.__validate_value_string(value)
        cls.__update_config()


class Configuration(object, metaclass=ConfigurationProperties):
    _use_config = False
    _save_config = False
    _config_file_path = '~/pyattck/config.yml'
    _data_path = '~/pyattck/data'
    _enterprise_attck_json = "https://raw.githubusercontent.com/mitre/cti/master/enterprise-attack/enterprise-attack.json"
    _pre_attck_json="https://raw.githubusercontent.com/mitre/cti/master/pre-attack/pre-attack.json"
    _mobile_attck_json="https://raw.githubusercontent.com/mitre/cti/master/mobile-attack/mobile-attack.json"
    _ics_attck_json="https://raw.githubusercontent.com/mitre/cti/master/ics-attack/ics-attack.json"
    _nist_controls_json="https://raw.githubusercontent.com/center-for-threat-informed-defense/attack-control-framework-mappings/master/frameworks/nist800-53-r4/stix/nist800-53-r4-controls.json"
    _generated_attck_json="https://github.com/swimlane/pyattck/blob/master/generated_attck_data.json?raw=True"
    _generated_nist_json="https://github.com/swimlane/pyattck/blob/master/attck_to_nist_controls.json?raw=True"
    _requests_kwargs = {}



from .attckobject import AttckObject
from ..utils.logo import Logo


class AttckActor(AttckObject):

    """Enterprise MITRE ATT&CK Actor object

    A child class of AttckObject

    Creates objects that are categorized as Mitre ATT&CK Enterprise
    Actors or Groups (e.g. APT1, APT32, etc.)

    You can also access external data properties. The following properties
    are generated using external data:

        1. country
        2. operations
        3. attribution_links
        4. known_tools
        5. targets
        6. additional_comments
        7. external_description

    You can retrieve the entire dataset using the `external_dataset` property.

    pyattck also enables you to retrieve or generate logos for the actor or
    group using the following properties:

        - ascii_logo - Generated ASCII logo based on the actor or groups name
        - image_logo - Generated ASCII logo based on a provided logo

    Example:
        You can iterate over an `actors` list and access specific properties and
        relationship properties.

        The following relationship properties are accessible:
                1. malwares
                2. tools
                3. techniques

            1. To iterate over an `actors` list, do the following:

            .. code-block:: python

               from pyattck import Attck

               attck = Attck()

               for actor in attck.enterprise.actors:
                   print(actor.id)
                   print(actor.name)
                   print(actor.alias)
                   print(actor.description)
                   # etc.

            2. To access relationship properties, do the following:

            .. code-block:: python

               from pyattck import Attck

               attck = Attck()

               for actor in attck.enterprise.actors:
                   print(actor.id)
                   print(actor.name)
                   print(actor.alias)
                   print(actor.description)

                   for malware in actor.malwares:
                       print(malware.name)
                       print(malware.description)
                       # etc.

    Arguments:
        attck_obj (json) -- Takes the raw Mitre ATT&CK Json object
        AttckObject (dict) -- Takes the Mitre ATT&CK Json object as a
                              kwargs values
    """

    __ATTCK_DATASETS = None

    def __init__(self, attck_obj = None, **kwargs):
        """
        This class represents a Actor (or group) as defined by the
        Enterprise MITRE ATT&CK framework.

        Keyword Arguments:
            attck_obj {json} -- A Enterprise MITRE ATT&CK Framework
                                json object (default: {None})

        Raises:
            GeneratedDatasetException: Raised an exception when unable
                                       to access or process the external
                                       generated dataset.
        """
        super(AttckActor, self).__init__(**kwargs)
        self.__attck_obj = attck_obj
        self.id = self._set_id(kwargs)
        self.created_by_ref = self._set_attribute(kwargs, 'created_by_ref')
        self.revoked = self._set_attribute(kwargs, 'revoked')
        self.name = self._set_attribute(kwargs, 'name')
        self.description = self._set_attribute(kwargs, 'description')
        self.external_reference = self._set_reference(kwargs)
        self.created = self._set_attribute(kwargs, 'created')
        self.modified = self._set_attribute(kwargs, 'modified')
        self.stix = self._set_attribute(kwargs, 'id')
        self.type = self._set_attribute(kwargs, 'type')
        self.wiki = self._set_wiki(kwargs)
        self.contributor = self._set_list_items(kwargs, 'x_mitre_contributors')
        self.set_relationships(self.__attck_obj)
        logo = Logo(self.name.strip().replace(' ','_').lower())
        self.ascii_logo = logo.get_ascii()
        self.image_logo = logo.get_image()
        self.external_dataset = self.__get_actors_dataset()

    def __get_actors_dataset(self):
        return_list = []
        countries = set()
        operations = set()
        attribution_links = set()
        known_tools = set()
        targets = set()
        additional_comments = set()
        external_description = set()
        for country in AttckObject.generated_attck_json.get('actors'):
            if country:
                for key,val in country.items():
                    for actor in val.get('actors'):
                        if self.name in actor.get('names') or hasattr(self, 'alias') and self.alias and any(x in actor.get('names') for x in self.alias):
                            countries.add(key)
                            if actor.get('operations'):
                                operations.update(actor.get('operations'))
                            if actor.get('links'):
                                attribution_links.update(actor.get('links'))
                            if actor.get('tools'):
                                known_tools.update(actor.get('tools'))
                            if actor.get('targets'):
                                targets.add(actor.get('targets'))
                            if actor.get('comment'):
                                additional_comments.update(actor.get('comment',[]))
                            if actor.get('description'):
                                external_description.update(actor.get('description'))
                            return_list.append(actor)
        self.country = list(countries)
        self.operations = list(operations)
        self.attribution_links = list(attribution_links)
        self.known_tools = list(known_tools)
        self.targets = list(targets)
        self.additional_comments = list(additional_comments)
        self.external_description = list(external_description)
        return return_list

    @property
    def malwares(self):
        """
        Returns all malware objects as a list that are documented as being
        used by an Actor or Group

        Returns:
            [list] -- A list of related malware objects defined within the
                      Enterprise MITRE ATT&CK Framework
        """
        from .malware import AttckMalware
        return_list = []
        item_dict = {}
        for item in self.__attck_obj['objects']:
            if 'type' in item:
                if item['type'] == 'malware':
                    item_dict[item['id']] = item
        if self._RELATIONSHIPS.get(self.stix):
            for item in self._RELATIONSHIPS[self.stix]:
                if item in item_dict:
                    return_list.append(AttckMalware(attck_obj=self.__attck_obj, **item_dict[item]))
        return return_list

    @property
    def tools(self):
        """
        Returns all tool object as a list that are documented as being
        used by an Actor or Group

        Returns:
            [list] -- A list of related tool objects defined within the
            Enterprise MITRE ATT&CK Framework
        """
        from .tools import AttckTools
        return_list = []
        item_dict = {}
        for item in self.__attck_obj['objects']:
            if 'type' in item:
                if item['type'] == 'tool':
                    item_dict[item['id']] = item
        if self._RELATIONSHIPS.get(self.stix):
            for item in self._RELATIONSHIPS.get(self.stix):
                if item in item_dict:
                    return_list.append(AttckTools(attck_obj=self.__attck_obj, **item_dict[item]))
        return return_list

    @property
    def techniques(self):
        """
        Returns all technique objects as a list that are documented as
        being used by an Actor or Group

        Returns:
            [list] -- A list of related technique objects defined within the
            Enterprise MITRE ATT&CK Framework
        """
        from .technique import AttckTechnique
        return_list = []
        item_dict = {}
        for item in self.__attck_obj['objects']:
            if 'type' in item:
                if item['type'] == 'attack-pattern':
                    item_dict[item['id']] = item
        if self._RELATIONSHIPS.get(self.stix):
            for item in self._RELATIONSHIPS[self.stix]:
                if item in item_dict:
                    return_list.append(AttckTechnique(attck_obj=self.__attck_obj, **item_dict[item]))
        return return_list



from setuptools import setup, find_packages

def parse_requirements(requirement_file):
    with open(requirement_file) as f:
        return f.readlines()

version = dict()
with open("./pyattck/utils/version.py") as fp:
    exec(fp.read(), version)


setup(
    name='pyattck',
    version=version['__version__'],
    packages=find_packages(exclude=['tests*']),
    license='MIT',
    description='A Python package to interact with the Mitre ATT&CK Frameworks',
    long_description=open('README.md').read(),
    long_description_content_type="text/markdown",
    install_requires=parse_requirements('./requirements.txt'),
    keywords=['att&ck', 'mitre', 'swimlane'],
    url='https://github.com/swimlane/pyattck',
    author='Swimlane',
    author_email='info@swimlane.com',
    python_requires='>=3.6, <4',
    package_data={
        'pyattck':  ['data/actors/*.png']
    },
    entry_points={
          'console_scripts': [
              'pyattck = pyattck.__main__:main'
          ]
    },
)



from .attckobject import AttckObject


class AttckTechnique(AttckObject):

    """Enterprise MITRE ATT&CK Technique object.

    A child class of AttckObject

    Creates objects which have been categorized as a technique used by attackers

    Each technique enables you to access the following properties on the object:

        1. command_list - A list of commands associated with a technique
        2. commands = A list of dictionary objects containing source, command,
                      and provided name associated with a technique
        3. queries = A list of dictionary objects containing product, query,
                     and name associated with a technique
        4. datasets = A list of raw datasets associated with a technique
        5. possible_detections = A list of raw datasets containing possible
                                 detection methods for a technique

    Example:
        You can iterate over an `techniques` list and access specific
        properties and relationship properties.

        The following relationship properties are accessible:
                1. tactics
                2. mitigations
                3. actors

            1. To iterate over an `techniques` list, do the following:

            .. code-block:: python

               from pyattck import Attck

               attck = Attck()

               for technique in attck.enterprise.techniques:
                   print(technique.id)
                   print(technique.name)
                   print(technique.alias)
                   print(technique.description)
                   # etc.

            2. To access relationship properties, do the following:

            .. code-block:: python

               from pyattck import Attck

               attck = Attck()

               for technique in attck.enterprise.techniques:
                   print(technique.id)
                   print(technique.name)
                   print(technique.alias)
                   print(technique.description)
                   # etc.

                   for malware in technique.malwares:
                       print(malware.name)
                       print(malware.description)
                       # etc.

                   # to get a count of controls for a technique do the following
                   print(len(technique.controls))

                   # below will print each controls properties & values
                   for control in technique.controls:
                       print(control.__dict__)

                   # below will print the id, name and description of a control
                   for control in technique.controls:
                       print(control.id)
                       print(control.name)
                       print(control.description)

    Arguments:
        attck_obj (json) -- Takes the raw MITRE ATT&CK Json object
        AttckObject (dict) -- Takes the MITRE ATT&CK Json object as a kwargs values
    """

    def __init__(self, attck_obj = None, **kwargs):
        """
        This class represents a Technique as defined by the
        Enterprise MITRE ATT&CK framework.

        Keyword Arguments:
            attck_obj {json} -- A Enterprise MITRE ATT&CK Framework
                                json object (default: {None})

        Raises:
            GeneratedDatasetException: Raised an exception when unable to
                                       access or process the external
                                       generated dataset.
        """
        super(AttckTechnique, self).__init__(**kwargs)
        self.__attck_obj = attck_obj
        self.created_by_reference = self._set_attribute(kwargs, 'created_by_ref')
        self.platforms = self._set_list_items(kwargs, 'x_mitre_platforms')
        self.permissions = self._set_list_items(kwargs, 'x_mitre_permissions_required')
        self.bypass = self._set_list_items(kwargs, 'x_mitre_defense_bypassed')
        self.effective_permissions = self._set_list_items(kwargs, 'x_mitre_effective_permissions')
        self.network = self._set_attribute(kwargs, 'x_mitre_network_requirements')
        self.remote = self._set_attribute(kwargs, 'x_mitre_remote_support')
        self.system_requirements = self._set_attribute(kwargs, 'x_mitre_system_requirements')
        self.detection = self._set_attribute(kwargs, 'x_mitre_detection')
        self.__data_sources = self._create_data_sources_dict(kwargs.get('x_mitre_data_sources'))
        self.created = self._set_attribute(kwargs, 'created')
        self.modified = self._set_attribute(kwargs, 'modified')
        self.__subtechniques = []
        self.wiki = self._set_wiki(kwargs)
        self.contributors = self._set_list_items(kwargs, 'x_mitre_contributors')
        self.revoked = self._set_attribute(kwargs, 'revoked')
        self.deprecated = self._set_attribute(kwargs, 'x_mitre_deprecated')
        self.subtechnique = False if self._set_attribute(kwargs, 'x_mitre_is_subtechnique') is None else True
        self.__subtechniques = []
        self.command_list = self.__get_filtered_dataset('command_list')
        self.commands = self.__get_filtered_dataset('commands')
        self.queries = self.__get_filtered_dataset('queries')
        self.datasets = self.__get_filtered_dataset('parsed_datasets')
        self.possible_detections = self.__get_filtered_dataset('possible_detections')
        self.tactics = kwargs
        self.set_relationships(self.__attck_obj)

    def __get_filtered_dataset(self, attribute_name):
        for item in AttckObject.generated_attck_json['techniques']:
            if item['technique_id'] == self.id:
                return item[attribute_name]

    def __get_subtechnique_id(self, obj):
        return obj.id

    @property
    def subtechniques(self):
        """
        Subtechniques are by default nested under their parent technique.
        To flatten all techniques onto the same level provide
        nested_subtechniques=False when instantiating the Attck object.

        Returns:
            [type]: [description]
        """
        return sorted(self.__subtechniques, key=self.__get_subtechnique_id)

    @subtechniques.setter
    def subtechniques(self, value):
        self.__subtechniques.append(value)

    @property
    def controls(self):
        """
        Returns all compliance control objects that are associated with a technique

        Returns:
            [list] -- A list of control objects defined within the
                      Enterprise MITRE ATT&CK Framework
        """
        from .control import AttckControl
        control_list = []
        if AttckObject.generated_nist_json.get(self.stix):
            for control in AttckObject.nist_controls_json:
                if control.get('id') in AttckObject.generated_nist_json[self.stix]:
                    control_list.append(AttckControl(**control))
        return control_list

    @property
    def data_components(self):
        from .datasource import AttckDataComponent
        return_list = []
        item_dict = {}
        for item in self.__attck_obj['objects']:
            if 'type' in item:
                if item['type'] == 'x-mitre-data-component':
                    item_dict[item['id']] = item
        if self._RELATIONSHIPS.get(self.stix):
            for item in self._RELATIONSHIPS[self.stix]:
                if item in item_dict:
                    return_list.append(AttckDataComponent(**item_dict[item]))
        return return_list 

    @property
    def data_sources(self):
        """
        Returns all data source objects that a technique belongs to

        Returns:
            [list] -- A list of data source objects defined within the
                      Enterprise MITRE ATT&CK Framework
        """
        from .datasource import AttckDataSource
        data_source_list = []
        for item in self.__attck_obj['objects']:
            if 'x-mitre-data-source' in item['type'] and self.__data_sources.get(item['name']):
                data_source_list.append(AttckDataSource(
                    attck_obj=self.__attck_obj, 
                    _data_component_filter=self.__data_sources[item['name']],
                    **item)
                )
        return data_source_list

    @property
    def tactics(self):
        """
        Returns all tactic object that a technique belongs to

        Returns:
            [list] -- A list of tactic objects defined within the
                      Enterprise MITRE ATT&CK Framework
        """
        from .tactic import AttckTactic
        tactic_list = []
        for item in self.__attck_obj['objects']:
            if 'x-mitre-tactic' in item['type']:
                for tact in self._tactic:
                    if str(tact).lower() == str(item['x_mitre_shortname']).lower():
                        tactic_list.append(AttckTactic(attck_obj=self.__attck_obj, **item))
        return tactic_list

    @tactics.setter
    def tactics(self, obj):
        """
        Sets the associated tactic/phase this technique is in

        Arguments:
            obj (dict) -- A MITRE ATT&CK Framework json object

        Returns:
            (string) -- Returns a string that sets the tactic/phase this technique is in. 
                        If there is no phase found, it will return 'no phase_name'
        """
        temp_list = []
        try:
            for phase in obj['kill_chain_phases']:
                temp_list.append(phase['phase_name'])
            self._tactic = temp_list
        except:
            self._tactic = ['no phase_name']
        
    @property
    def mitigations(self):
        """
        Returns all mitigation objects that a technique is associated with

        Returns:
            [list] -- A list of mitigation objects defined within the
                      Enterprise MITRE ATT&CK Framework
        """
        from .mitigation import AttckMitigation
        return_list = []
        item_dict = {}
        for item in self.__attck_obj['objects']:
            if 'type' in item:
                if item['type'] == 'course-of-action':
                    item_dict[item['id']] = item
        if self._RELATIONSHIPS.get(self.stix):
            for item in self._RELATIONSHIPS[self.stix]:
                if item in item_dict:
                    return_list.append(AttckMitigation(attck_obj=self.__attck_obj, **item_dict[item]))
        return return_list 

    @property
    def actors(self):
        """
        Returns all actor objects that use a technique

        Returns:
            [list] -- A list of actor objects defined within the
                      Enterprise MITRE ATT&CK Framework
        """
        from .actor import AttckActor
        return_list = []
        item_dict = {}
        for item in self.__attck_obj['objects']:
            if 'type' in item:
                if item['type'] == 'intrusion-set':
                    item_dict[item['id']] = item
        if self._RELATIONSHIPS.get(self.stix):
            for item in self._RELATIONSHIPS[self.stix]:
                if item in item_dict:
                    return_list.append(AttckActor(attck_obj=self.__attck_obj, **item_dict[item]))
        return return_list



from .attckobject import AttckObject


class AttckMitigation(AttckObject):

    """Enterprise MITRE ATT&CK Mitigation object.

    A child class of AttckObject

    Creates objects which have been categorized as potential mitigations

    Example:
        You can iterate over a `mitigations` list and access specific
        properties and relationship properties.

        The following relationship properties are accessible:
                1. techniques

            1. To iterate over an `mitigations` list, do the following:

            .. code-block:: python

               from pyattck import Attck

               attck = Attck()

               for mitigation in attck.enterprise.mitigations:
                   print(mitigation.id)
                   print(mitigation.name)
                   print(mitigation.description)
                   # etc.

            2. To access relationship properties, do the following:

            .. code-block:: python

               from pyattck import Attck

               attck = Attck()

               for mitigation in attck.enterprise.mitigations:
                   print(mitigation.id)
                   print(mitigation.name)
                   print(mitigation.description)
                   # etc.

                   for technique in mitigation.techniques:
                       print(technique.name)
                       print(technique.description)
                       # etc.

    Arguments:s
        attck_obj (json) -- Takes the raw Mitre ATT&CK Json object
        AttckObject (dict) -- Takes the Mitre ATT&CK Json object as a kwargs values
    """

    def __init__(self, attck_obj = None, **kwargs):
        """
        This class represents mitigation guidance as defined by the
        Enterprise MITRE ATT&CK framework.

        Keyword Arguments:
            attck_obj {json} -- A Enterprise MITRE ATT&CK Framework json object (default: {None})
        """
        super(AttckMitigation, self).__init__(**kwargs)
        self.__attck_obj = attck_obj
        self.created_by_ref = self._set_attribute(kwargs, 'created_by_ref')
        self.id = self._set_id(kwargs)
        self.name = self._set_attribute(kwargs, 'name')
        self.description = self._set_attribute(kwargs, 'description')
        self.external_reference = self._set_reference(kwargs)
        self.created = self._set_attribute(kwargs, 'created')
        self.modified = self._set_attribute(kwargs, 'modified')
        self.stix = self._set_attribute(kwargs, 'id')
        self.type = self._set_attribute(kwargs, 'type')
        self.wiki = self._set_wiki(kwargs)
        self.contributor = self._set_attribute(kwargs, 'contributor')
        self.set_relationships(self.__attck_obj)

    @property
    def techniques(self):
        """
        Returns all technique objects as a list that are associated with this
        mitigation advice from the Enterprise MITRE ATT&CK Framework

        Returns:
            [list] -- A list of related technique objects defined within the
                      Enterprise MITRE ATT&CK Framework for a mitigation object
        """
        from .technique import AttckTechnique
        return_list = []
        item_dict = {}
        for item in self.__attck_obj['objects']:
            if 'type' in item:
                if item['type'] == 'attack-pattern':
                    item_dict[item['id']] = item
        if self._RELATIONSHIPS.get(self.stix):
            for item in self._RELATIONSHIPS[self.stix]:
                if item in item_dict:
                    return_list.append(AttckTechnique(attck_obj=self.__attck_obj, **item_dict[item]))
        return return_list



from .technique import MobileAttckTechnique
from .actor import MobileAttckActor
from .malware import MobileAttckMalware
from .tools import MobileAttckTools
from .mitigation import MobileAttckMitigation
from .tactic import MobileAttckTactic
from ..configuration import Configuration


class MobileAttck(object):

    """Mobile MITRE ATT&CK interface.

    This class creates an interface to all data points in the
    MITRE Mobile ATT&CK framework.

    This interface enables you to retrieve all properties within
    each item in the MITRE Mobile ATT&CK Framework.

    The following categorical items can be accessed using this class:

        1. Tactics (Tactics are the phases defined by MITRE Mobile ATT&CK)
        2. Techniques (Techniques are the individual actions which can accomplish
           a tactic)
        3. Mitigations (Mitigations are recommendations to prevent or protect against
           a technique)
        4. Actors (Actors or Groups are identified malicious actors/groups
           which have been identified and documented by MITRE & third-parties)
        5. Tools (Tools are software used to perform techniques)
        6. Malwares (Malwares are specific pieces of malware used by actors
           (or in general) to accomplish a technique)

    Additionally, as of pyattck 2.0.0 you can now access additional datasets related
    to a technique. These datasets are
    [documented here](https://github.com/swimlane/pyattck/blob/master/generateattcks/README.md).

    Each technique enables you to access the following properties on the object:

        1. command_list - A list of commands associated with a technique
        2. commands = A list of dictionary objects containing source, command, and
                      provided name associated with a technique
        3. queries = A list of dictionary objects containing product, query, and name
                     associated with a technique
        4. datasets = A list of raw datasets associated with a technique
        5. possible_detections = A list of raw datasets containing possible detection
                                 methods for a technique

    Each Actor object (if available) enables you to access the following properties on
    the object:

        1. country
        2. operations
        3. attribution_links
        4. known_tools
        5. targets
        6. additional_comments
        7. external_description

    You can retrieve the entire dataset using the `external_dataset` property on
    a `actor` object. pyattck also enables you to retrieve or generate logos for
    the actor or group using the following properties:

        - ascii_logo - Generated ASCII logo based on the actor or groups name

    Each Tools object (if available) enables you to access the following properties on the object:

        1. additional_names
        2. attribution_links
        3. additional_comments
        4. family

    You can retrieve the entire dataset using the `external_dataset` property
    on a `tool` object. You can also access external data properties from the
    C2 Matrix project. The following properties are generated using C2 Matrix
    external data:

        - HTTP
        - Implementation
        - Custom Profile
        - DomainFront
        - Multi-User
        - SMB
        - Kill Date
        - macOS
        - GitHub
        - Key Exchange
        - Chaining
        - Price
        - TCP
        - Proxy Aware
        - HTTP3
        - HTTP2
        - Date
        - Evaluator
        - Working Hours
        - Slack
        - FTP
        - Version Reviewed
        - Logging
        - Name
        - License
        - Windows
        - Stego
        - Notes
        - Server
        - Actively Maint.
        - Dashboard
        - DNS
        - Popular Site
        - ICMP
        - IMAP
        - DoH
        - Jitter
        - How-To
        - ATT&CK Mapping
        - Kali
        - Twitter
        - MAPI
        - Site
        - Agent
        - API
        - UI
        - Linux

    You can retrieve the entire dataset using the `c2_data` property.

    Example:
        Once an Attck object is instantiated, you can access each object type as a
        list of objects (e.g. techniques, tactics, actors, etc.)

        You can iterate over each object list and access specific properties and
        relationship properties of each.
    
        The following relationship properties are accessible:

        1. Actors
            1. Tools used by the Actor or Group
            2. Malware used by the Actor or Group
            3. Techniques this Actor or Group uses
        2. Malwares
            1. Actor or Group(s) using this malware
            2. Techniques this malware is used with
        3. Mitigations
            1. Techniques related to a specific set of mitigation suggestions
        4. Tactics
            1. Techniques found in a specific Tactic (phase)
        5. Techniques
            1. Tactics a technique is found in
            2. Mitigation suggestions for a given technique
            3. Actor or Group(s) identified as using this technique
        6. Tools
            1. Techniques that the specified tool is used within
            2. Actor or Group(s) using a specified tool

        1. To iterate over a list, do the following:

        .. code-block:: python

            from pyattck import Attck

            attck = Attck()
            
            for technique in attck.mobile.techniques:
                print(technique.id)
                print(technique.name)
                print(technique.description)
                # etc.
            for mitigation in attck.mobile.mitigations:
                print(mitigation.id)
                print(mitigation.name)
                print(mitigation.description)
                # etc.

        2. To access relationship properties, do the following:

        .. code-block:: python

            from pyattck import Attck

            attck = Attck()

            for technique in attck.mobile.techniques:
                print(technique.id)
                print(technique.name)
                print(technique.description)
                # etc.

                for actor in technique.actors:
                    print(actor.id)
                    print(actor.name)
                    print(actor.description)
                    # etc.

            for mitigation in attck.mobile.mitigations:
                print(mitigation.id)
                print(mitigation.name)
                print(mitigation.description)
                # etc.

                for technique in mitigation.techniques:
                    print(technique.name)
                    print(technique.description)
                    # etc.
    """

    __ENTERPRISE_GENERATED_DATA_JSON = None
    __mobile_attck_json = Configuration.get_data('mobile_attck_json')
    __tactics = []
    __techniques = []
    __mitigations = []
    __actors = []
    __tools = []
    __malwares = []

    @property
    def actors(self):
        """
        Creates MobileAttckActor objects

        Returns:
            (MobileAttckActor) -- (Returns a list of MobileAttckActor objects)
        """
        if not self.__actors:
            for group in self.__mobile_attck_json['objects']:
                if group['type'] == 'intrusion-set':
                    self.__actors.append(MobileAttckActor(mobile_attck_obj=self.__mobile_attck_json, **group))
        return self.__actors

    @property
    def tactics(self):
        """
        Creates MobileAttckTactic objects

        Returns:
            (MobileAttckTactic) -- (Returns a list of MobileAttckTactic objects)
        """
        if not self.__tactics:
            for tactic in self.__mobile_attck_json['objects']:
                if tactic['type'] == 'x-mitre-tactic':
                    self.__tactics.append(MobileAttckTactic(mobile_attck_obj=self.__mobile_attck_json, **tactic))
        return self.__tactics

    @property
    def mitigations(self):
        """
        Creates MobileAttckMitigation objects

        Returns:
            (MobileAttckMitigation) -- (Returns a list of MobileAttckMitigation objects)
        """
        if not self.__mitigations:
            for mitigation in self.__mobile_attck_json['objects']:
                if mitigation['type'] == 'course-of-action':
                    self.__mitigations.append(MobileAttckMitigation(mobile_attck_obj=self.__mobile_attck_json, **mitigation))
        return self.__mitigations

    @property
    def tools(self):
        """
        Creates MobileAttckTools objects

        Returns:
            (MobileAttckTools) -- Returns a list of MobileAttckTools objects
        """
        if not self.__tools:
            for tools in self.__mobile_attck_json['objects']:
                if tools['type'] == 'tool':
                    self.__tools.append(MobileAttckTools(mobile_attck_obj=self.__mobile_attck_json, **tools))
        return self.__tools

    @property
    def malwares(self):
        """
        Creates MobileAttckMalware objects

        Returns:
            (MobileAttckMalware) -- Returns a list of MobileAttckMalware objects
        """
        if not self.__malwares:
            for malware in self.__mobile_attck_json['objects']:
                if malware['type'] == 'malware':
                    self.__malwares.append(MobileAttckMalware(mobile_attck_obj=self.__mobile_attck_json, **malware))
        return self.__malwares

    @property
    def techniques(self):
        """
        Creates MobileAttckTechnique objects

        Returns:
            (MobileAttckTechnique) -- Returns a list of MobileAttckTechnique objects
        """
        if not self.__techniques:
            for technique in self.__mobile_attck_json["objects"]:
                if technique['type'] == 'attack-pattern':
                    self.__techniques.append(MobileAttckTechnique(mobile_attck_obj=self.__mobile_attck_json, **technique))
        return self.__techniques
